#!/usr/bin/env python
r"""
Common functions used in scipy's curve_fit or minimize parameter estimations.

For all fit functions, it defines the functions in two forms (ex. of 3
params):

.. code-block:: python

   func(x, p1, p2, p3)
   func_p(x, p) with p[0:3]

The first form can be used, for example, with
`scipy.optimize.curve_fit` (ex. function f1x=a+b/x):

.. code-block:: python

   p, cov = scipy.optimize.curve_fit(functions.f1x, x, y, p0=[p0, p1])

It also defines two cost functions along with the fit functions, one
with the absolute sum, one with the squared sum of the deviations:

.. code-block:: python

   cost_func  = sum(abs(obs-func))
   cost2_func = sum((obs-func)**2)

These cost functions can be used, for example, with
`scipy.optimize.minimize`:

.. code-block:: python

   p = scipy.optimize.minimize(
           jams.functions.cost2_f1x, np.array([p1, p2]), args=(x, y),
           method='Nelder-Mead', options={'disp': False})

Note the different argument orders:
    1. `curvefit` needs `f(x, *p)` with the independent variable as the first
       argument and the parameters to fit as separate remaining arguments.
    2. `minimize` is a general minimiser with respect to the first argument,
       i.e. `func(p, *args)`.

The module provides also two common cost functions (absolute and
squared deviations) where any function in the form `func(x, p)` can be
used as second argument:

.. code-block:: python

   cost_abs(p, func, x, y)
   cost_square(p, func, x, y)

This means, for example `cost_f1x(p, x, y)` is the same as
`cost_abs(p, functions.f1x_p, x, y)`. For example:

.. code-block:: python

   p = scipy.optimize.minimize(
           jams.functions.cost_abs, np.array([p1, p2]),
           args=(functions.f1x_p, x, y), method='Nelder-Mead',
           options={'disp': False})

There are the following functions in four forms. They have the name in the
first column. The second form has a `_p` appended to the name. The cost
functions have `cost_` or `cost2_` prepended to the name, for example
:func:`gauss`, :func:`gauss_p`, :func:`cost_gauss`, :func:`cost2_gauss`:

.. list-table::
   :widths: 15 50
   :header-rows: 1

   * - Function - number of paramters
     - Description

   * - :func:`arrhenius`
     - Arrhenius temperature dependence of biochemical rates:
       :math:`\exp((T-TC25)*E/(T25*R*(T+T0)))`

   * - :func:`f1x`
     - General 1/x function: :math:`a + b/x`

   * - :func:`fexp`
     - General exponential function: :math:`a + b * \exp(c*x)`

   * - :func:`gauss`
     - Gauss function:
       :math:`1/(\sigma*\sqrt(2*\pi)) * \exp(-(x-\mu)^2/(2*\sigma^2))`

   * - :func:`lasslop`
     - NEE of Lasslop et al. (2010)

   * - :func:`line0`
     - Straight line: :math:`a*x`

   * - :func:`line`
     - Straight line: :math:`a + b*x`

   * - :func:`lloyd_fix`
     - Heterotrophic respiration of Lloyd & Taylor (1994)

   * - :func:`lloyd_only_rref`
     - Heterotrophic respiration of Lloyd & Taylor (1994) with fixed E0

   * - :func:`logistic`
     - Logistic function: :math:`a/(1+\exp(-b(x-c)))`

   * - :func:`logistic_offset`
     - Logistic function with offset: :math:`a/(1+\exp(-b(x-c))) + d`

   * - :func:`logistic2_offset`
     - Double logistic function with offset
       :math:`L1/(1+\exp(-k1(x-x01))) - L2/(1+\exp(-k2(x-x02))) + a`

   * - :func:`poly`
     - General polynomial: :math:`c_0 + c_1*x + c_2*x^2 + ... + c_n*x^n`

   * - :func:`sabx`
     - sqrt(f1x), i.e. general sqrt(1/x) function: :math:`\sqrt(a + b/x)`

   * - :func:`see`
     - Sequential Elementary Effects fitting function: :math:`a*(x-b)^c`

This module was written by Matthias Cuntz while at Department of
Computational Hydrosystems, Helmholtz Centre for Environmental
Research - UFZ, Leipzig, Germany, and continued while at Institut
National de Recherche pour l'Agriculture, l'Alimentation et
l'Environnement (INRAE), Nancy, France.

:copyright: Copyright 2012-2022 Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

Functions:

.. autosummary::
   cost_abs
   cost_square
   arrhenius
   arrhenius_p
   cost_arrhenius
   cost2_arrhenius
   f1x
   f1x_p
   cost_f1x
   cost2_f1x
   fexp
   fexp_p
   cost_fexp
   cost2_fexp
   gauss
   gauss_p
   cost_gauss
   cost2_gauss
   lasslop
   lasslop_p
   cost_lasslop
   cost2_lasslop
   line
   line_p
   cost_line
   cost2_line
   line0
   line0_p
   cost_line0
   cost2_line0
   lloyd_fix
   lloyd_fix_p
   cost_lloyd_fix
   cost2_lloyd_fix
   lloyd_only_rref
   lloyd_only_rref_p
   cost_lloyd_only_rref
   cost2_lloyd_only_rref
   sabx
   sabx_p
   cost_sabx
   cost2_sabx
   poly
   poly_p
   cost_poly
   cost2_poly
   cost_logistic
   cost2_logistic
   cost_logistic_offset
   cost2_logistic_offset
   cost_logistic2_offset
   cost2_logistic2_offset
   see
   see_p
   cost_see
   cost2_see

History
    * Written Dec 2012 by Matthias Cuntz (mc (at) macu (dot) de)
    * Ported to Python 3, Feb 2013, Matthias Cuntz
    * Added general cost functions `cost_abs` and `cost_square`,
      May 2013, Matthias Cuntz
    * Added line0, Feb 2014, Matthias Cuntz
    * Removed `multiline_p`, May 2020, Matthias Cuntz
    * Changed to Sphinx docstring and numpydoc, May 2020, Matthias Cuntz
    * More consistent docstrings, Jan 2022, Matthias Cuntz

"""
import numpy as np
from .logistic_function import logistic_p, logistic_offset_p
from .logistic_function import logistic2_offset_p
from ..const import T0, T25, R
# try:         # import package
#     from .logistic_function import logistic_p, logistic_offset_p
#     from .logistic_function import logistic2_offset_p
#     from ..const import T0, T25, R
# except:
#     try:     # e.g. python nee2gpp.py
#         from functions.logistic_function import logistic_p, logistic_offset_p
#         from functions.logistic_function import logistic2_offset_p
#         from const import T0, T25, R
#     except:  # python fit_functions.py
#         from logistic_function import logistic_p, logistic_offset_p
#         from logistic_function import logistic2_offset_p
#         T0  = 273.15     # Celcius <-> Kelvin [K]
#         T25 = 298.15     # Standard ambient temperature [K]
#         R   = 8.3144621  # Ideal gas constant [J K^-1 mol^-1]


__all__ = ['cost_abs', 'cost_square',
           'arrhenius', 'arrhenius_p', 'cost_arrhenius', 'cost2_arrhenius',
           'f1x', 'f1x_p', 'cost_f1x', 'cost2_f1x',
           'fexp', 'fexp_p', 'cost_fexp', 'cost2_fexp',
           'gauss', 'gauss_p', 'cost_gauss', 'cost2_gauss',
           'lasslop', 'lasslop_p', 'cost_lasslop', 'cost2_lasslop',
           'line', 'line_p', 'cost_line', 'cost2_line',
           'line0', 'line0_p', 'cost_line0', 'cost2_line0',
           'lloyd_fix', 'lloyd_fix_p', 'cost_lloyd_fix', 'cost2_lloyd_fix',
           'lloyd_only_rref', 'lloyd_only_rref_p', 'cost_lloyd_only_rref',
           'cost2_lloyd_only_rref',
           'sabx', 'sabx_p', 'cost_sabx', 'cost2_sabx',
           'poly', 'poly_p', 'cost_poly', 'cost2_poly',
           'cost_logistic', 'cost2_logistic',
           'cost_logistic_offset', 'cost2_logistic_offset',
           'cost_logistic2_offset', 'cost2_logistic2_offset',
           'see', 'see_p', 'cost_see', 'cost2_see']


# -----------------------------------------------------------
# general cost functions
def cost_abs(p, func, x, y):
    """
    General cost function for robust optimising `func(x, p)` vs `y` with sum
    of absolute deviations.

    Parameters
    ----------
    p : iterable of floats
        parameters
    func : callable
        `fun(x,p) -> float`
    x : float or array_like of floats
        independent variable
    y : float or array_like of floats
        dependent variable, observations

    Returns
    -------
    float
        sum of absolute deviations
    """
    return np.sum(np.abs(y - func(x, p)))


def cost_square(p, func, x, y):
    """
    General cost function for optimising `func(x, p)` vs `y` with sum of
    square deviations.

    Parameters
    ----------
    p : iterable of floats
        parameters
    func : callable
        `fun(x,p) -> float`
    x : float or array_like of floats
        independent variable
    y : float or array_like of floats
        dependent variable, observations

    Returns
    -------
    float
        sum of squared deviations
    """
    return np.sum((y - func(x, p))**2)


# -----------------------------------------------------------
# arrhenius
def arrhenius(T, E):
    """
    Arrhenius temperature dependence of rates

    Parameters
    ----------
    T : float or array_like of floats
        temperature [degC]
    E : float
        activation energy [J]

    Returns
    -------
    float
        function value(s)
    """
    return np.exp( ( T - (T25 - T0)) * E / (T25 * R * (T + T0)) )


def arrhenius_p(T, p):
    """
    Arrhenius temperature dependence of rates

    Parameters
    ----------
    T : float or array_like of floats
        temperature [degC]
    p : iterable
        `p[0]` = activation energy [J]

    Returns
    -------
    float
        function value(s)
    """
    return np.exp((T - (T25 - T0)) * p[0] / (T25 * R * (T + T0)))


def cost_arrhenius(p, T, rate):
    """
    Sum of absolute deviations of obs and arrhenius function

    Parameters
    ----------
    p : iterable of floats
        `p[0]` = activation energy [J]
    x : float or array_like of floats
        independent variable
    y : float or array_like of floats
        dependent variable, observations

    Returns
    -------
    float
        sum of absolute deviations
    """
    return np.sum(np.abs(rate - arrhenius_p(T, p)))


def cost2_arrhenius(p, T, rate):
    """
    Sum of squared deviations of obs and arrhenius function

    Parameters
    ----------
    p : iterable of floats
        `p[0]` = activation energy [J]
    x : float or array_like of floats
        independent variable
    y : float or array_like of floats
        dependent variable, observations

    Returns
    -------
    float
        sum of squared deviations
    """
    return np.sum((rate - arrhenius_p(T, p))**2)


# -----------------------------------------------------------
# a+b/x
def f1x(x, a, b):
    """
    General 1/x function: :math:`a + b/x`

    Parameters
    ----------
    x : float or array_like of floats
        independent variable
    a : float
        first parameter
    b : float
        second parameter

    Returns
    -------
    float
        function value(s)
    """
    return a + b / x


def f1x_p(x, p):
    """
    General 1/x function: :math:`a + b/x`

    Parameters
    ----------
    x : float or array_like of floats
        independent variable
    p : iterable of floats
        parameters (`len(p)=2`)
          - `p[0]` = a
          - `p[1]` = b

    Returns
    -------
    float
        function value(s)
    """
    return p[0] + p[1] / x


def cost_f1x(p, x, y):
    """
    Sum of absolute deviations of obs and general 1/x function: :math:`a + b/x`

    Parameters
    ----------
    p : iterable of floats
        parameters (`len(p)=2`)
          - `p[0]` = a
          - `p[1]` = b
    x : float or array_like of floats
        independent variable
    y : float or array_like of floats
        dependent variable, observations

    Returns
    -------
    float
        sum of absolute deviations
    """
    return np.sum(np.abs(y - f1x_p(x, p)))


def cost2_f1x(p, x, y):
    """
    Sum of squared deviations of obs and general 1/x function: :math:`a + b/x`

    Parameters
    ----------
    p : iterable of floats
        parameters (`len(p)=2`)
          - `p[0]` = a
          - `p[1]` = b
    x : float or array_like of floats
        independent variable
    y : float or array_like of floats
        dependent variable, observations

    Returns
    -------
    float
        sum of squared deviations
    """
    return np.sum((y - f1x_p(x, p))**2)


# -----------------------------------------------------------
# a+b*exp(c*x)
def fexp(x, a, b, c):
    """
    General exponential function: :math:`a + b * exp(c*x)`

    Parameters
    ----------
    x : float or array_like of floats
        independent variable
    a : float
        first parameter
    b : float
        second parameter
    c : float
        third parameter

    Returns
    -------
    float
        function value(s)
    """
    return a + b * np.exp(c * x)


def fexp_p(x, p):
    """
    General exponential function: :math:`a + b * exp(c*x)`

    Parameters
    ----------
    x : float or array_like of floats
        independent variable
    p : iterable of floats
        parameters (`len(p)=3`)
          - `p[0]` = a
          - `p[1]` = b
          - `p[2]` = c

    Returns
    -------
    float
        function value(s)
    """
    return p[0] + p[1] * np.exp(p[2] * x)


def cost_fexp(p, x, y):
    """
    Sum of absolute deviations of obs and general exponential function:
    :math:`a + b * exp(c*x)`

    Parameters
    ----------
    p : iterable of floats
        parameters (`len(p)=3`)
          - `p[0]` = a
          - `p[1]` = b
          - `p[2]` = c
    x : float or array_like of floats
        independent variable
    y : float or array_like of floats
        dependent variable, observations

    Returns
    -------
    float
        sum of absolute deviations
    """
    return np.sum(np.abs(y - fexp_p(x, p)))


def cost2_fexp(p, x, y):
    """
    Sum of squared deviations of obs and general exponential function:
    :math:`a + b * exp(c*x)`

    Parameters
    ----------
    p : iterable of floats
        parameters (`len(p)=3`)
          - `p[0]` = a
          - `p[1]` = b
          - `p[2]` = c
    x : float or array_like of floats
        independent variable
    y : float or array_like of floats
        dependent variable, observations

    Returns
    -------
    float
        sum of squared deviations
    """
    return np.sum((y - fexp_p(x, p))**2)


# -----------------------------------------------------------
# Gauss: 1/(sig*sqrt(2*pi)) *exp(-(x-mu)**2/(2*sig**2))
def gauss(x, mu, sig):
    """
    Gauss function:
    :math:`1 / (sqrt(2*pi)*sig) * exp( -(x-mu)**2 / (2*sig**2) )`

    Parameters
    ----------
    x : float or array_like of floats
        independent variable
    mu : float
        mean
    sig : float
        width

    Returns
    -------
    float
        function value(s)
    """
    return np.exp(-(x - mu)**2 / (2. * sig**2)) / (sig * np.sqrt(2. * np.pi))


def gauss_p(x, p):
    """
    Gauss function:
    :math:`1 / (sqrt(2*pi)*sig) * exp( -(x-mu)**2 / (2*sig**2) )`

    Parameters
    ----------
    x : float or array_like of floats
        independent variable
    p : iterable of floats
        parameters (`len(p)=2`)
          - `p[0]` = mean mu
          - `p[1]` = width sig

    Returns
    -------
    float
        function value(s)
    """
    return (np.exp(-(x - p[0])**2 / (2. * p[1]**2)) /
            (p[1] * np.sqrt(2. * np.pi)))


def cost_gauss(p, x, y):
    """
    Sum of absolute deviations of obs and Gauss function:
    :math:`1 / (sqrt(2*pi)*sig) * exp( -(x-mu)**2 / (2*sig**2) )`

    Parameters
    ----------
    p : iterable of floats
        parameters (`len(p)=2`)
          - `p[0]` = mean mu
          - `p[1]` = width sig
    x : float or array_like of floats
        independent variable
    y : float or array_like of floats
        dependent variable, observations

    Returns
    -------
    float
        sum of absolute deviations
    """
    return np.sum(np.abs(y - gauss_p(x, p)))


def cost2_gauss(p, x, y):
    """
    Sum of squared deviations of obs and Gauss function:
    :math:`1 / (sqrt(2*pi)*sig) * exp( -(x-mu)**2 / (2*sig**2) )`

    Parameters
    ----------
    p : iterable of floats
        parameters (`len(p)=2`)
          - `p[0]` = mean mu
          - `p[1]` = width sig
    x : float or array_like of floats
        independent variable
    y : float or array_like of floats
        dependent variable, observations

    Returns
    -------
    float
        sum of squared deviations
    """
    return np.sum((y - gauss_p(x, p))**2)


# -----------------------------------------------------------
# lasslop
def lasslop(Rg, et, VPD, alpha, beta0, k, Rref):
    """
    NEE of Lasslop et al (2010)

    Lasslop et al. (2010) is the rectangular, hyperbolic light-response
    of NEE as by Falge et al. (2001), where the respiration is calculated
    with Lloyd & Taylor (1994), and the maximum canopy uptake rate at
    light saturation decreases exponentially with VPD as in Koerner (1995).

    Parameters
    ----------
    Rg : float or array_like of floats
        Global radiation [W m-2]
    et : float or array_like of floats
        Exponential in Lloyd & Taylor: np.exp(E0*(1./(Tref-T0)-1./(T-T0))) []
    VPD : float or array_like of floats
        Vapour Pressure Deficit [Pa]
    alpha : float
        Light use efficiency, i.e. initial slope of light response curve
        [umol(C) J-1]
    beta0 : float
        Maximum CO2 uptake rate at VPD0=10 hPa [umol(C) m-2 s-1]
    k : float
        e-folding of exponential decrease of maximum CO2 uptake with
        VPD increase [Pa-1]
    Rref : float
        Respiration at Tref (10 degC) [umol(C) m-2 s-1]

    Returns
    -------
    float
        net ecosystem exchange [umol(CO2) m-2 s-1]
    """
    # Lloyd & Taylor (1994)
    gamma = Rref * et
    # Koerner (1995)
    VPD0  = 1000.  # 10 hPa
    kk    = np.clip(-k * (VPD - VPD0), -600., 600.)
    beta  = np.where(VPD > VPD0, beta0 * np.exp(kk), beta0)
    return -alpha * beta * Rg / (alpha * Rg + beta) + gamma


def lasslop_p(Rg, et, VPD, p):
    """
    NEE of Lasslop et al (2010)

    Lasslop et al. (2010) is the rectangular, hyperbolic light-response
    of NEE as by Falge et al. (2001), where the respiration is calculated
    with Lloyd & Taylor (1994), and the maximum canopy uptake rate at
    light saturation decreases exponentially with VPD as in Koerner (1995).

    Parameters
    ----------
    Rg : float or array_like of floats
        Global radiation [W m-2]
    et : float or array_like of floats
        Exponential in Lloyd & Taylor: np.exp(E0*(1./(Tref-T0)-1./(T-T0))) []
    VPD : float or array_like of floats
        Vapour Pressure Deficit [Pa]
    p : iterable of floats
        parameters (`len(p)=4`)
          - `p[0]` = Light use efficiency, i.e. initial slope of light response
            curve [umol(C) J-1]
          - `p[1]` = Maximum CO2 uptake rate at VPD0=10 hPa [umol(C) m-2 s-1]
          - `p[2]` = e-folding of exponential decrease of maximum CO2 uptake
            with VPD increase [Pa-1]
          - `p[3]` = Respiration at Tref (10 degC) [umol(C) m-2 s-1]

    Returns
    -------
    float
        net ecosystem exchange [umol(CO2) m-2 s-1]
    """
    # Lloyd & Taylor (1994)
    gamma = p[3] * et
    # Koerner (1995)
    VPD0  = 1000.  # 10 hPa
    kk    = np.clip(-p[2] * (VPD - VPD0), -600., 600.)
    beta  = np.where(VPD > VPD0, p[1] * np.exp(kk), p[1])
    return -p[0] * beta * Rg / (p[0] * Rg + beta) + gamma


def cost_lasslop(p, Rg, et, VPD, NEE):
    """
    Sum of absolute deviations of obs and Lasslop et al (2010)

    Parameters
    ----------
    p : iterable of floats
        parameters (`len(p)=4`)
          - `p[0]` = Light use efficiency, i.e. initial slope of light response
            curve [umol(C) J-1]
          - `p[1]` = Maximum CO2 uptake rate at VPD0=10 hPa [umol(C) m-2 s-1]
          - `p[2]` = e-folding of exponential decrease of maximum CO2 uptake
            with VPD increase [Pa-1]
          - `p[3]` = Respiration at Tref (10 degC) [umol(C) m-2 s-1]
    Rg : float or array_like of floats
        Global radiation [W m-2]
    et : float or array_like of floats
        Exponential in Lloyd & Taylor: np.exp(E0*(1./(Tref-T0)-1./(T-T0))) []
    VPD : float or array_like of floats
        Vapour Pressure Deficit [Pa]
    NEE : float or array_like of floats
        Observed net ecosystem exchange [umol(CO2) m-2 s-1]

    Returns
    -------
    float
        sum of absolute deviations
    """
    return np.sum(np.abs(NEE - lasslop(Rg, et, VPD, p[0], p[1], p[2], p[3])))


def cost2_lasslop(p, Rg, et, VPD, NEE):
    """
    Sum of squared deviations of obs and Lasslop et al (2010)

    Parameters
    ----------
    p : iterable of floats
        parameters (`len(p)=4`)
          - `p[0]` = Light use efficiency, i.e. initial slope of light response
            curve [umol(C) J-1]
          - `p[1]` = Maximum CO2 uptake rate at VPD0=10 hPa [umol(C) m-2 s-1]
          - `p[2]` = e-folding of exponential decrease of maximum CO2 uptake
            with VPD increase [Pa-1]
          - `p[3]` = Respiration at Tref (10 degC) [umol(C) m-2 s-1]
    Rg : float or array_like of floats
        Global radiation [W m-2]
    et : float or array_like of floats
        Exponential in Lloyd & Taylor: np.exp(E0*(1./(Tref-T0)-1./(T-T0))) []
    VPD : float or array_like of floats
        Vapour Pressure Deficit [Pa]
    NEE : float or array_like of floats
        Observed net ecosystem exchange [umol(CO2) m-2 s-1]

    Returns
    -------
    float
        sum of squared deviations
    """
    return np.sum((NEE - lasslop(Rg, et, VPD, p[0], p[1], p[2], p[3]))**2)


# -----------------------------------------------------------
# a+b*x
def line(x, a, b):
    """
    Straight line: :math:`a + b*x`

    Parameters
    ----------
    x : float or array_like of floats
        independent variable
    a : float
        first parameter
    b : float
        second parameter

    Returns
    -------
    float
        function value(s)
    """
    return a + b * x


def line_p(x, p):
    """
    Straight line: :math:`a + b*x`

    Parameters
    ----------
    x : float or array_like of floats
        independent variable
    p : iterable of floats
        parameters (`len(p)=2`)
          - `p[0]` = a
          - `p[1]` = b

    Returns
    -------
    float
        function value(s)
    """
    return p[0] + p[1] * x


def cost_line(p, x, y):
    """
    Sum of absolute deviations of obs and straight line: :math:`a + b*x`

    Parameters
    ----------
    p : iterable of floats
        parameters (`len(p)=2`)
          - `p[0]` = a
          - `p[1]` = b
    x : float or array_like of floats
        independent variable
    y : float or array_like of floats
        dependent variable, observations

    Returns
    -------
    float
        sum of absolute deviations
    """
    return np.sum(np.abs(y - line_p(x, p)))


def cost2_line(p, x, y):
    """
    Sum of squared deviations of obs and straight line: :math:`a + b*x`

    Parameters
    ----------
    p : iterable of floats
        parameters (`len(p)=2`)
          - `p[0]` = a
          - `p[1]` = b
    x : float or array_like of floats
        independent variable
    y : float or array_like of floats
        dependent variable, observations

    Returns
    -------
    float
        sum of squared deviations
    """
    return np.sum((y - line_p(x, p))**2)


# -----------------------------------------------------------
# b*x
def line0(x, a):
    """
    Straight line through origin: :math:`a*x`

    Parameters
    ----------
    x : float or array_like of floats
        independent variable
    a : float
        first parameter

    Returns
    -------
    float
        function value(s)
    """
    return a * x


def line0_p(x, p):
    """
    Straight line through origin: :math:`a*x`

    Parameters
    ----------
    x : float or array_like of floats
        independent variable
    p : iterable of floats
        `p[0]` = a

    Returns
    -------
    float
        function value(s)
    """
    return p * x


def cost_line0(p, x, y):
    """
    Sum of absolute deviations of obs and straight line through origin:
    :math:`a*x`

    Parameters
    ----------
    p : iterable of floats
        `p[0]` = a
    x : float or array_like of floats
        independent variable
    y : float or array_like of floats
        dependent variable, observations

    Returns
    -------
    float
        sum of absolute deviations
    """
    return np.sum(np.abs(y - line0_p(x, p)))


def cost2_line0(p, x, y):
    """
    Sum of squared deviations of obs and straight line through origin:
    :math:`a*x`

    Parameters
    ----------
    p : iterable of floats
        `p[0]` = a
    x : float or array_like of floats
        independent variable
    y : float or array_like of floats
        dependent variable, observations

    Returns
    -------
    float
        sum of squared deviations
    """
    return np.sum((y - line0_p(x, p))**2)


# -----------------------------------------------------------
# lloyd_fix
def lloyd_fix(T, Rref, E0):
    """
    Soil respiration of Lloyd & Taylor (1994)

    Lloyd & Taylor (1994) Arrhenius type with T0=-46.02 degC and Tref=10 degC

    Parameters
    ----------
    T : float or array_like of floats
        Temperature [K]
    Rref : float
        Respiration at Tref=10 degC [umol(C) m-2 s-1]
    E0 : float
        Activation energy [K]

    Returns
    -------
    float
        Respiration [umol(C) m-2 s-1]
    """
    Tref = 283.15  # 10 [degC]
    T0   = 227.13  # -46.02 [degC]
    return Rref * np.exp(E0 * (1. / (Tref - T0) - 1. / (T - T0)))


def lloyd_fix_p(T, p):
    """
    Soil respiration of Lloyd & Taylor (1994)

    Lloyd & Taylor (1994) Arrhenius type with T0=-46.02 degC and Tref=10 degC

    Parameters
    ----------
    T : float or array_like of floats
        Temperature [K]
    p : iterable of floats
        parameters (`len(p)=2`)
          - `p[0]` = Respiration at Tref=10 degC [umol(C) m-2 s-1]
          - `p[1]` = Activation energy [K]

    Returns
    -------
    float
        Respiration [umol(C) m-2 s-1]
    """
    Tref = 283.15  # 10 [degC]
    T0   = 227.13  # -46.02 [degC]
    return p[0] * np.exp(p[1] * (1. / (Tref - T0) - 1. / (T - T0)))


def cost_lloyd_fix(p, T, resp):
    """
    Sum of absolute deviations of obs and Lloyd & Taylor (1994) Arrhenius type.

    Parameters
    ----------
    p : iterable of floats
        parameters (`len(p)=2`)
          - `p[0]` = Respiration at Tref=10 degC [umol(C) m-2 s-1]
          - `p[1]` = Activation energy [K]
    T : float or array_like of floats
        Temperature [K]
    resp : float or array_like of floats
        Observed respiration [umol(C) m-2 s-1]

    Returns
    -------
    float
        sum of absolute deviations
    """
    return np.sum(np.abs(resp - lloyd_fix_p(T, p)))


def cost2_lloyd_fix(p, T, resp):
    """
    Sum of squared deviations of obs and Lloyd & Taylor (1994) Arrhenius type.

    Parameters
    ----------
    p : iterable of floats
        parameters (`len(p)=2`)
          - `p[0]` = Respiration at Tref=10 degC [umol(C) m-2 s-1]
          - `p[1]` = Activation energy [K]
    T : float or array_like of floats
        Temperature [K]
    resp : float or array_like of floats
        Observed respiration [umol(C) m-2 s-1]

    Returns
    -------
    float
        sum of squared deviations
    """
    return np.sum((resp - lloyd_fix_p(T, p))**2)


# -----------------------------------------------------------
# lloyd_only_rref
def lloyd_only_rref(et, Rref):
    """
    Soil respiration of Lloyd & Taylor (1994) with fix E0

    If E0 is know in Lloyd & Taylor (1994) then one can calc
    the exponential term outside the routine and the fitting
    becomes linear. One could also use functions.line0.

    Parameters
    ----------
    et : float or array_like of floats
        exp-term in Lloyd & Taylor
    Rref : float
        Respiration at Tref=10 degC [umol(C) m-2 s-1]

    Returns
    -------
    float
        Respiration [umol(C) m-2 s-1]
    """
    return Rref * et


def lloyd_only_rref_p(et, p):
    """
    Soil respiration of Lloyd & Taylor (1994) with fix E0

    If E0 is know in Lloyd & Taylor (1994) then one can calc
    the exponential term outside the routine and the fitting
    becomes linear. One could also use functions.line0.

    Parameters
    ----------
    et : float or array_like of floats
        exp-term in Lloyd & Taylor
    p : iterable of floats
        `p[0]` = Respiration at Tref=10 degC [umol(C) m-2 s-1]

    Returns
    -------
    float
        Respiration [umol(C) m-2 s-1]
    """
    return p[0] * et


def cost_lloyd_only_rref(p, et, resp):
    """
    Sum of absolute deviations of obs and Lloyd & Taylor with fixed E0

    Parameters
    ----------
    p : iterable of floats
        `p[0]` = Respiration at Tref=10 degC [umol(C) m-2 s-1]
    et : float or array_like of floats
        exp-term in Lloyd & Taylor
    resp : float or array_like of floats
        Observed respiration [umol(C) m-2 s-1]

    Returns
    -------
    float
        sum of absolute deviations
    """
    return np.sum(np.abs(resp - lloyd_only_rref_p(et, p)))


def cost2_lloyd_only_rref(p, et, resp):
    """
    Sum of squared deviations of obs and Lloyd & Taylor with fixed E0

    Parameters
    ----------
    p : iterable of floats
        `p[0]` = Respiration at Tref=10 degC [umol(C) m-2 s-1]
    et : float or array_like of floats
        exp-term in Lloyd & Taylor
    resp : float or array_like of floats
        Observed respiration [umol(C) m-2 s-1]

    Returns
    -------
    float
        sum of squared deviations
    """
    return np.sum((resp - lloyd_only_rref_p(et, p))**2)


# -----------------------------------------------------------
# sqrt(a + b/x) - theoretical form of Jackknife-after-bootstrap

def sabx(x, a, b):
    """
    Square root of general 1/x function: :math:`sqrt(a + b/x)`

    Parameters
    ----------
    x : float or array_like of floats
        independent variable
    a : float
        first parameter
    b : float
        second parameter

    Returns
    -------
    float
        function value(s)
    """
    return np.sqrt(a + b / x)


def sabx_p(x, p):
    """
    Square root of general 1/x function: :math:`sqrt(a + b/x)`

    Parameters
    ----------
    x : float or array_like of floats
        independent variable
    p : iterable of floats
        parameters (`len(p)=2`)
          - `p[0]` = a
          - `p[1]` = b

    Returns
    -------
    float
        function value(s)
    """
    return np.sqrt(p[0] + p[1] / x)


def cost_sabx(p, x, y):
    """
    Sum of absolute deviations of obs and square root of general 1/x function:
    :math:`sqrt(a + b/x)`

    Parameters
    ----------
    p : iterable of floats
        parameters (`len(p)=2`)
          - `p[0]` = a
          - `p[1]` = b
    x : float or array_like of floats
        independent variable
    y : float or array_like of floats
        dependent variable, observations

    Returns
    -------
    float
        sum of absolute deviations
    """
    return np.sum(np.abs(y - sabx_p(x, p)))


def cost2_sabx(p, x, y):
    """
    Sum of squared deviations of obs and square root of general 1/x function:
    :math:`sqrt(a + b/x)`

    Parameters
    ----------
    p : iterable of floats
        parameters (`len(p)=2`)
          - `p[0]` = a
          - `p[1]` = b
    x : float or array_like of floats
        independent variable
    y : float or array_like of floats
        dependent variable, observations

    Returns
    -------
    float
        sum of squared deviations
    """
    return np.sum((y - sabx_p(x, p))**2)


# -----------------------------------------------------------
# c0 + c1*x + c2*x**2 + ... + cn*x**n
def poly(x, *args):
    """
    General polynomial: :math:`c0 + c1*x + c2*x**2 + ... + cn*x**n`

    Parameters
    ----------
    x : float or array_like of floats
        independent variable
    *args : float
        parameters `len(args)=n+1`

    Returns
    -------
    float
        function value(s)
    """
    return np.polynomial.polynomial.polyval(x, list(args))


def poly_p(x, p):
    """
    General polynomial: :math:`c0 + c1*x + c2*x**2 + ... + cn*x**n`

    Parameters
    ----------
    x : float or array_like of floats
        independent variable
    p : iterable of floats
        parameters (`len(p)=n+1`)

    Returns
    -------
    float
        function value(s)
    """
    return np.polynomial.polynomial.polyval(x, p)


def cost_poly(p, x, y):
    """
    Sum of absolute deviations of obs and general polynomial:
    :math:`c0 + c1*x + c2*x**2 + ... + cn*x**n`

    Parameters
    ----------
    p : iterable of floats
        parameters (`len(p)=n+1`)
    x : float or array_like of floats
        independent variable
    y : float or array_like of floats
        dependent variable, observations

    Returns
    -------
    float
        sum of absolute deviations
    """
    return np.sum(np.abs(y - poly_p(x, p)))


def cost2_poly(p, x, y):
    """
    Sum of squared deviations of obs and general polynomial:
    :math:`c0 + c1*x + c2*x**2 + ... + cn*x**n`

    Parameters
    ----------
    p : iterable of floats
        parameters (`len(p)=n+1`)
    x : float or array_like of floats
        independent variable
    y : float or array_like of floats
        dependent variable, observations

    Returns
    -------
    float
        sum of squared deviations
    """
    return np.sum((y - poly_p(x, p))**2)


# -----------------------------------------------------------
# a/(1+exp(-b(x-c))) - logistic function
def cost_logistic(p, x, y):
    """
    Sum of absolute deviations of obs and logistic function
    :math:`L/(1+exp(-k(x-x0)))`

    Parameters
    ----------
    p : iterable of floats
        parameters (`len(p)=3`)
          - `p[0]` = L  = Maximum of logistic function
          - `p[1]` = k  = Steepness of logistic function
          - `p[2]` = x0 = Inflection point of logistic function
    x : float or array_like of floats
        independent variable
    y : float or array_like of floats
        dependent variable, observations

    Returns
    -------
    float
        sum of absolute deviations
    """
    return np.sum(np.abs(y - logistic_p(x, p)))


def cost2_logistic(p, x, y):
    """
    Sum of squared deviations of obs and logistic function
    :math:`L/(1+exp(-k(x-x0)))`

    Parameters
    ----------
    p : iterable of floats
        parameters (`len(p)=3`)
          - `p[0]` = L  = Maximum of logistic function
          - `p[1]` = k  = Steepness of logistic function
          - `p[2]` = x0 = Inflection point of logistic function
    x : float or array_like of floats
        independent variable
    y : float or array_like of floats
        dependent variable, observations

    Returns
    -------
    float
        sum of squared deviations
    """
    return np.sum((y - logistic_p(x, p))**2)


# -----------------------------------------------------------
# a/(1+exp(-b(x-c))) + d - logistic function with offset
def cost_logistic_offset(p, x, y):
    """
    Sum of absolute deviations of obs and logistic function 1/x function:
    :math:`L/(1+exp(-k(x-x0))) + a`

    Parameters
    ----------
    p : iterable of floats
        parameters (`len(p)=4`)
          - `p[0]` = L  = Maximum of logistic function
          - `p[1]` = k  = Steepness of logistic function
          - `p[2]` = x0 = Inflection point of logistic function
          - `p[3]` = a  = Offset of logistic function
    x : float or array_like of floats
        independent variable
    y : float or array_like of floats
        dependent variable, observations

    Returns
    -------
    float
        sum of absolute deviations
    """
    return np.sum(np.abs(y - logistic_offset_p(x, p)))


def cost2_logistic_offset(p, x, y):
    """
    Sum of squared deviations of obs and logistic function 1/x function:
    :math:`L/(1+exp(-k(x-x0))) + a`

    Parameters
    ----------
    p : iterable of floats
        parameters (`len(p)=4`)
          - `p[0]` = L  = Maximum of logistic function
          - `p[1]` = k  = Steepness of logistic function
          - `p[2]` = x0 = Inflection point of logistic function
          - `p[3]` = a  = Offset of logistic function
    x : float or array_like of floats
        independent variable
    y : float or array_like of floats
        dependent variable, observations

    Returns
    -------
    float
        sum of squared deviations
    """
    return np.sum((y - logistic_offset_p(x, p))**2)


# -----------------------------------------------------------
# double logistic function with offset
# L1/(1+exp(-k1(x-x01))) - L2/(1+exp(-k2(x-x02))) + a2
def cost_logistic2_offset(p, x, y):
    """
    Sum of absolute deviations of obs and double logistic function with offset:
    :math:`L1/(1+exp(-k1(x-x01))) - L2/(1+exp(-k2(x-x02))) + a`

    Parameters
    ----------
    p : iterable of floats
        parameters (`len(p)=7`)
          - `p[0]` = L1  = Maximum of first logistic function
          - `p[1]` = k1  = Steepness of first logistic function
          - `p[2]` = x01 = Inflection point of first logistic function
          - `p[3]` = L2  = Maximum of second logistic function
          - `p[4]` = k2  = Steepness of second logistic function
          - `p[5]` = x02 = Inflection point of second logistic function
          - `p[6]` = a  = Offset of double logistic function
    x : float or array_like of floats
        independent variable
    y : float or array_like of floats
        dependent variable, observations

    Returns
    -------
    float
        sum of absolute deviations
    """
    return np.sum(np.abs(y - logistic2_offset_p(x, p)))


def cost2_logistic2_offset(p, x, y):
    """
    Sum of squared deviations of obs and double logistic function with offset:
    :math:`L1/(1+exp(-k1(x-x01))) - L2/(1+exp(-k2(x-x02))) + a`

    Parameters
    ----------
    p : iterable of floats
        parameters (`len(p)=7`)
          - `p[0]` = L1  = Maximum of first logistic function
          - `p[1]` = k1  = Steepness of first logistic function
          - `p[2]` = x01 = Inflection point of first logistic function
          - `p[3]` = L2  = Maximum of second logistic function
          - `p[4]` = k2  = Steepness of second logistic function
          - `p[5]` = x02 = Inflection point of second logistic function
          - `p[6]` = a  = Offset of double logistic function
    x : float or array_like of floats
        independent variable
    y : float or array_like of floats
        dependent variable, observations

    Returns
    -------
    float
        sum of squared deviations
    """
    return np.sum((y - logistic2_offset_p(x, p))**2)


# -----------------------------------------------------------
# a*(x-b)**c - Sequential Elementary Effects fitting function
def see(x, a, b, c):
    """
    Fit function of Sequential Elementary Effects: :math:`a * (x-b)**c`

    Parameters
    ----------
    x : float or array_like of floats
        independent variable
    a : float
        first parameter
    b : float
        second parameter
    c : float
        third parameter

    Returns
    -------
    float
        function value(s)
    """
    return np.where((x - b) < 0., 0., a * (x - b)**c)


def see_p(x, p):
    """
    Fit function of Sequential Elementary Effects: :math:`a * (x-b)**c`

    Parameters
    ----------
    x : float or array_like of floats
        independent variable
    p : iterable of floats
        parameters (`len(p)=3`)
          - `p[0]` = a
          - `p[1]` = b
          - `p[2]` = c

    Returns
    -------
    float
        function value(s)
    """
    return np.where((x - p[1]) < 0., 0., p[0] * (x - p[1])**p[2])


def cost_see(p, x, y):
    """
    Sum of absolute deviations of obs and fit function of Sequential Elementary
    Effects: :math:`a * (x-b)**c`

    Parameters
    ----------
    p : iterable of floats
        parameters (`len(p)=3`)
          - `p[0]` = a
          - `p[1]` = b
          - `p[2]` = c
    x : float or array_like of floats
        independent variable
    y : float or array_like of floats
        dependent variable, observations

    Returns
    -------
    float
        sum of absolute deviations
    """
    return np.sum(np.abs(y - see_p(x, p)))


def cost2_see(p, x, y):
    """
    Sum of squared deviations of obs and fit function of Sequential Elementary
    Effects: :math:`a * (x-b)**c`

    Parameters
    ----------
    p : iterable of floats
        parameters (`len(p)=3`)
          - `p[0]` = a
          - `p[1]` = b
          - `p[2]` = c
    x : float or array_like of floats
        independent variable
    y : float or array_like of floats
        dependent variable, observations

    Returns
    -------
    float
        sum of squared deviations
    """
    return np.sum((y - see_p(x, p))**2)


# -----------------------------------------------------------

if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # Rref = 1.0
    # E0   = 126.
    # T    = 293.15
    # resp = 2.0
    # print(lloyd_fix(T, Rref, E0))
    # #1.40590910521
    # print(lloyd_fix_p(T, [Rref, E0]))
    # #1.40590910521
    # print(cost_lloyd_fix([Rref, E0], T, resp))
    # #0.59409089479
    # print(cost2_lloyd_fix([Rref, E0], T, resp))
    # #0.352943991272

    # print(poly(T,2,1))
    # #295.15
    # print(poly_p(T,[2,1]))
    # #295.15



#!/usr/bin/env python
'''
Download ERA5 or ERA5-Land data from Copernicus Climate Data Store.

It uses the CDS API. You have to have an account on ECMWF's Climate Data Store
(CDS) or Atmospheric Data Store (ADS). Follow the instructions in the section
'Download ERA5 family data through the CDS API' on:
https://confluence.ecmwf.int/display/CKB/How+to+download+ERA5

If override=False (default), the script checks if the data is already available
in the local download directory (path).
It expects files with the same naming convention than its own, i.e.
     path + '/' + 'era5_'+area.replace('/','_')+'_{:04d}.nc'.format(year)
or
     path + '/' + 'era5_'+area.replace('/','_')+'_{:04d}-{:04d}.nc'.format(yearstart, yearend)
Filenames can also end on .nc?.
It checks if the area and years are included in a file in the download
directory by checking ONLY filenames.

Be aware that the request is processed once it is queued even if you abort
this script. Queued request can be deleted after login at:
   https://cds.climate.copernicus.eu/cdsapp#!/yourrequests
You can (re-)download the data from this site later as well.

Written  Matthias Cuntz, Jan 2019 - from get_era_interim.py
Modified Stephan Thober, Mar 2020 - added era5-land capability
         Matthias Cuntz, Jun 2020
             - return correct file list and not the list of projected filenames
               for chosen area
             - input variable list with MuSICA variables as default
             - finalised era5-land capability
         Matthias Cuntz, Feb 2021 - bug in single point -> lat1 > lat2
         Matthias Cuntz, Feb 2021 - added grib format
                                  - default download era5-land in grib format
         Matthias Cuntz, Sep 2021 - no default area: -a must be given.
         Matthias Cuntz & Jerome Ogee, Sep 2022
             - respect latency of provision of ERA5 and ERA5-Land for
               current year
         Matthias Cuntz, Oct 2022
             - shift area by 0.01 so account for sites on grid borders of ERA
             - allow filenames starting with era5-land and era5land


--------------------------------------------------------
usage: get_era5.py [-h] [-a area] [-f format] [-o] [-p path]
                   [-r reanalysis_model] [-v variables] [-y years]

        Download ERA5 or ERA5-Land data from Copernicus Climate Data Store
        https://climate.copernicus.eu/climate-data-store.


optional arguments:
  -h, --help            show this help message and exit
  -a area, --area area  area format as either lat,lon or
                        NorthLat/WestLon/SouthLat/EastLon
                        (default: global 90/-180/-90/180).
  -f format, --format format
                        Output format netcdf or grib.
                        (default: netcdf if era5, grib if era5-land).
  -o, --override        Do not check that output file already exists that
                        includes request. Override existing output file
                        (default: False).
  -p path, --path path  Output directory (default: current directory ".").
  -r reanalysis_model, --reanalyis-model reanalysis_model
                        Reanalyis model to download, either: era5, era5-land or
                        era5land (default: era5)
  -v variables, --variables variables
                        Comma-separated variable list var1,var2,...
                        (default: forcing variables of ecosystem model MuSICA:
                        10m_u_component_of_wind,10m_v_component_of_wind,
                        2m_temperature,2m_dewpoint_temperature,total_precipitation,
                        snowfall,surface_pressure,
                        surface_solar_radiation_downwards,
                        surface_thermal_radiation_downwards).
  -y years, --years years
                        years format is startyear,endyear
                        (default: 1979,current-1 for ERA5 and
                        1981,current-1 for ERA5-Land.)


Examples
--------
    # Hesse
    python get_era5.py -r ERA5-Land -v 2m_temperature,2m_dewpoint_temperature \
                       -a 48.6742166667,7.06461666667 -y 1995,2017 -p era
    # Tumbarumba
    python get_era5.py -r ERA5-Land -v 10m_u_component_of_wind,10m_v_component_of_wind,2m_temperature,2m_dewpoint_temperature,total_precipitation,snowfall,surface_pressure,surface_solar_radiation_downwards,surface_thermal_radiation_downwards --area=-35.7833302,148.0166666 -y 2016,2018 -p era


--------------------------------------------------------

Script was originally adapted from CDS Web API
    https://cds.climate.copernicus.eu/cdsapp#!/dataset/reanalysis-era5-single-levels?tab=form
Further help comes from
    https://confluence.ecmwf.int/display/CKB/C3S+ERA5%3A+Web+API+to+CDS+API
'''
import os
import glob
import datetime as dt


__all__ = ['get_era5']


# --------------------------------------------------------------------
# Retrieval function
#

def _get_era5_single_level5(variables, date, time, area, target,
                            grid=None, reanalysis_model='era5',
                            output_format='netcdf'):
    """
    ToDo
    """
    import cdsapi

    # check reanalysis mode
    retrieve_name = {'era5': 'reanalysis-era5-single-levels',
                     'era5land': 'reanalysis-era5-land',
                     'era5-land': 'reanalysis-era5-land'}
    if reanalysis_model not in retrieve_name.keys():
        estr = 'Value of variable reanalysis_model'
        estr = estr + ' {:s} is not in retrieve_name.keys()'.format(
            reanalysis_model)
        raise ValueError(estr)

    # Disable: InsecureRequestWarning: Unverified HTTPS request is being made.
    #          Adding certificate verification is strongly advised.
    try:
        import urllib3
        urllib3.disable_warnings()
    except:
        import requests
        from requests.packages.urllib3.exceptions import InsecureRequestWarning
        requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

    server = cdsapi.Client()

    request = {
        'format': output_format,  # grib or netcdf
        'variable': variables,    # ['10m_v_component_of_wind', ...]
        'date': date,             # "2010-01-01", "2010-01-01/2015-12-31"
        'time': time,             # "[00:00, 01:00, 02:00, ..., 23:00]
        'area': area,             # North, West, South, East within
                                  # "90/-180/-90/180" or [90, -180, -90, 180]
    }

    if reanalysis_model == 'era5':
        request['product_type'] = 'reanalysis'

    if grid:
        # 'grid': [1.0,1.0]
        # Latitude/longitude grid: east-west (longitude)
        # and north-south resolution (latitude). Default: 0.25 x 0.25
        request['grid'] = grid

    # print('Request: ', request)
    server.retrieve(retrieve_name[reanalysis_model], request, target)

    return target


def _get_era5_single_level5_datestarget(variables, time, area, datestarget,
                                        grid=None, reanalysis_model='era5',
                                        output_format='netcdf'):
    """
    Wrapper function for `_get_era5_single_level5` to parallelise with
    Python's `multiprocessing.Pool`.
    """
    date, target = datestarget
    return _get_era5_single_level5(variables, date, time, area, target,
                                   grid, reanalysis_model=reanalysis_model,
                                   output_format=output_format)


# --------------------------------------------------------------------
# Main routine
#

def get_era5(vars=['10m_u_component_of_wind', '10m_v_component_of_wind',
                   '2m_temperature', '2m_dewpoint_temperature',
                   'total_precipitation', 'snowfall', 'surface_pressure',
                   'surface_solar_radiation_downwards',
                   'surface_thermal_radiation_downwards'],
             area='90/-180/-90/180', years=None, path='.',
             override=False, reanalysis_model='era5',
             output_format=''):
    """
    Download ERA5 or ERA5-Land data from Copernicus Climate Data Store.

    If override=False, checks if the data is already available in
    the download directory.
    It expects files with the same naming convention than its own, i.e.
        path + '/' + 'era5<land>_'+area.replace('/','_')+'_{:04d}.nc'.format(year)
    or
        path + '/' + 'era5<land>_'+area.replace('/','_')+'_{:04d}-{:04d}.nc'.format(yearstart, yearend)
    Filenames can also end on .nc?.
    It checks if the area and years are included in a file by checking
    ONLY filenames.

    Be aware that the request is processed once it is queued even if you
    abort this function.
    Queued request can be deleted after login at:
       https://cds.climate.copernicus.eu/cdsapp#!/yourrequests
    You can (re-)download the data from this site later as well.

    Parameters
    ----------
    vars : list of str, optional
        List of names of variables to retrieve.
        Default are the input variables for the ecosystem model MuSICA:

        `vars=['10m_u_component_of_wind', '10m_v_component_of_wind',
               'total_precipitation',
               'snowfall', '2m_temperature', '2m_dewpoint_temperature',
               'surface_pressure',
               'surface_solar_radiation_downwards',
               'surface_thermal_radiation_downwards']`

        Look up variable names for ERA5 by selecting variables and examining
        the API (Show API request):
        https://cds.climate.copernicus.eu/cdsapp#!/dataset/reanalysis-era5-single-levels?tab=form

        and for ERA5-Land:
        https://cds.climate.copernicus.eu/cdsapp#!/dataset/reanalysis-era5-land?tab=form
    area : string, optional
        Area. Can be one point given as lat,lon
        or box as NorthLat/WestLon/SouthLat/EastLon.
        Default is the globe: '90/-180/-90/180'.

        Min and Max of Lat and Lon of box must be different by
        at least 0.25 degree for ERA5 and 0.1 degree for ERA5-Land.
    years : scalar or iterable of len(2), optional
        Year or year range (default: [1979,current_year-1] for ERA5
        and [1981,current_year-1] for ERA5-Land)
    path : string, optional
        Output path (default: '.')
    override : bool, optional
        If True, check if data is already available in path (default: False).

        Expects file with the naming convention

            path + '/' + 'era5<land>_'+area.replace('/','_')+'_{:04d}.nc'.format(year)

        or

            path + '/' + 'era5<land>_'+area.replace('/','_')+'_{:04d}-{:04d}.nc'.format(yearstart, yearend)

        Checks if `area` and `years` are included in a file in path by
        checking ONLY filenames.
    reanalysis_model : string, optional
        Reanalyis model to download. Can be era5, era5-land or era5land
        (default: 'era5').
        Output filenames will be adapted accordingly.
    output_format : string, optional
        File format of output file.
        Default is 'netcdf' if `reanalysis_model=='era5'` and
        'grib' if `reanalysis_model=='era5-land'.
        Output filenames will be suffixed by '.nc' or '.grb', respectively.

        Model output at ECMWF is stored in grib format. There are limitations
        on the conversion to netCDF using the current ECMWF infrastructure.
        One gets errors like 'One or more variable sizes violate format
        constraints.':
        https://confluence.ecmwf.int/display/CKB/Common+Error+Messages+for+CDS+Requests
        Download in grib format in this case and use the climate data operators
        'cdo' to convert to netCDF format:

        Check the file content to know latitude, longitude and year:

            `cdo -t ecmwf sinfov gribfile`

        Construct the netcdf name, e.g. era5-land_lat1_lon1_lat2_lon2_year.nc
        Convert to compressed netCDF4 format:

            `cdo -f nc4 -z zip -t ecmwf copy gribfile netcdffile`

        If you want to have lowercase variable names, this is in bash:

            `gribfile=ecmwf_output.grb`
            `# cdo -t ecmwf sinfov ${gribfile}`
            `year=$(echo $(cdo -s showyear ${gribfile}))`
            `netcdffile=era5-land_75_-15_28_45_${year}.nc`
            `vars=$(cdo -s -t ecmwf showname ${gribfile})`
            `copt="" ; for i in ${vars} ; do copt="${copt} -chname,${i},$(echo ${i} | tr A-Z a-z)" ; done`
            `cdo -L -f nc4 -z zip -t ecmwf ${copt} ${gribfile} ${netcdffile}`

    Returns
    -------
    list
        Returns filenames of the output files in path, either the newly written
        files or the files that contain the output requested (netcdf example):

            [ path + '/' + 'era5<land>_' + area.replace('/','_')
            + '_{:04d}.nc'.format(yy) for yy in years ]

    Warnings
    --------
    Existing files will only be checked by filename not by content.

    Examples
    --------
    area  = '48/7/47/8'
    years = (1995,2017)
    ofile = get_era5(area=area, years=years, path='.')
    file1 = 'era5_'+area.replace('/','_')+'_{:04d}.nc'.format(years[0])
    if file1 != ofile[0]: print(
        'Returned filename not recognised: ', ofile, ' Expected: ', file1)
    import os
    if not os.path.exists(ofile): print('No ofile: ', ofile)
    if not os.path.exists(file1): print('No file1: ', file1)

    License
    -------
    This file is part of the JAMS Python package,
    distributed under the MIT License.

    Copyright (c) 2019-2022 Matthias Cuntz - mc (at) macu (dot) de


    History
    -------
    Written  Matthias Cuntz, Jan 2019 - from get_era_interim.py
    Modified Matthias Cuntz, Dec 2019
                 - default area (global) was not working:
                   used == instead of =
             Stephan Thober, Mar 2020
                 - added optional reanalysis_model argument to download
                   era5land
             Matthias Cuntz, Jun 2020
                 - allow name era5-land
                 - return correct file list and not the list of projected
                   filenames for chosen area
                 - input variable list with MuSICA variables as default
                 - finalised era5-land capability
                 - use numpydoc format
             Matthias Cuntz, Feb 2021 - bug in single point -> lat1 > lat2
             Matthias Cuntz, Feb 2021
                 - added grib format
                 - download era5-land in grib format
             Matthias Cuntz & Jerome Ogee, Sep 2022
                 - respect latency of provision of ERA5 and ERA5-Land for
                   current year
             Matthias Cuntz, Oct 2022
                 - shift area by 0.01 so account for sites on grid borders of
                   ERA
                 - allow filenames starting with era5-land and era5land

    """
    # Check parameters
    # reanalysis model
    rmodel = reanalysis_model.lower()
    estr = 'Reanalysis model must be era5, era5-land or era5land.'
    assert rmodel in ['era5', 'era5-land', 'era5land'], estr
    if rmodel == 'era5':
        resolution = 0.25
        minyear    = 1979
        if not output_format:
            output_format = 'netcdf'
        # era5 needs days that are already present in the product.
        # It has a latency of about 5 days.
        latency = 10
    else:
        # rmodel     = 'era5land'
        resolution = 0.1
        minyear    = 1981
        if not output_format:
            output_format = 'grib'
        # era5-land has a latency of about 2-3 months.
        # But it seems to accept days that do not exist in the product yet.
        latency = 0
    output_format = output_format.lower()
    if output_format == 'netcdf':
        suffix = '.nc'
    elif output_format == 'grib':
        suffix = '.grb'
    else:
        estr  = 'Output format must be netcdf or grib. Given: '
        estr += output_format
        raise ValueError(estr)

    # area
    if '/' in area:
        sarea = area.split('/')
        estr = 'area format is lat,lon or NorthLat/WestLon/SouthLat/EastLon.'
        assert len(sarea) == 4, estr
        estr = 'area must be in 90/-180/-90/180'
        assert float(sarea[0]) <= 90., estr
        assert float(sarea[1]) >= -180., estr
        assert float(sarea[2]) >= -90., estr
        assert float(sarea[3]) <= 180., estr
        estr = 'area format is lat,lon or NorthLat/WestLon/SouthLat/EastLon.'
        assert float(sarea[0]) >= float(sarea[2]) + resolution, estr
        estr = 'area format is lat,lon or NorthLat/WestLon/SouthLat/EastLon.'
        assert float(sarea[1]) + resolution <= float(sarea[3]), estr
    else:
        estr = 'area format is lat,lon or NorthLat/WestLon/SouthLat/EastLon.'
        assert ',' in area, estr
        sarea = area.split(',')
        estr = 'area format is lat,lon or NorthLat/WestLon/SouthLat/EastLon.'
        assert len(sarea) == 2, estr
        lat, lon = sarea
        # single point does not work. Needs to encompass an actual grid point.
        # take slightly more to the left and bottom to take lower value when
        # exactly on grid border
        area = (str(float(lat) + resolution / 2 - 0.01) + '/'
                + str(float(lon) - resolution / 2 + 0.01) + '/'
                + str(float(lat) - resolution / 2 + 0.01) + '/'
                + str(float(lon) + resolution / 2 - 0.01))

    # years
    if years:
        estr = 'years must be scaler or iterable with two elements.'
        assert len(years) <= 2, estr
        if len(years) == 1:
            yearstart = yearend = years
        else:
            yearstart, yearend = [ int(i) for i in years ]
        estr = 'Start year must be greater'
        estr = estr + ' {:d} for reanalysis model {:s}.'.format(
            minyear, reanalysis_model)
        assert yearstart >= minyear, estr
    else:
        import time as ptime
        yearstart = minyear
        curryear  = int(ptime.asctime().split()[-1])
        yearend   = curryear - 1

    # Make output directory
    if not os.path.exists(path):
        os.makedirs(path)

    # Single year output files
    hasyrs     = list(range(yearstart, yearend + 1))
    targetera5 = [ path + '/' + rmodel + '_' + area.replace('/', '_')
                   + '_{:04d}'.format(yy) + suffix for yy in hasyrs ]
    targetyrs  = targetera5[:]

    # Check existing files
    files = glob.glob(path + '/*.nc*') + glob.glob(path + '/*.gr*b*')
    files = [ ff for ff in files
              if os.path.basename(ff).startswith(rmodel + '_') ]
    if (len(files) > 0) and (not override):
        ilat1, ilon1, ilat2, ilon2 = [ float(i) for i in area.split('/') ]
        for ff in files:
            if len(hasyrs) > 0:
                fs = ff.split('_')
                lat1 = float(fs[-5])
                lon1 = float(fs[-4])
                lat2 = float(fs[-3])
                lon2 = float(fs[-2])
                if ( (ilat1 <= lat1) and (ilat2 >= lat2) and
                     (ilon1 >= lon1) and (ilon2 <= lon2) ):
                    fs  = fs[-1]
                    yrs = fs[:fs.rfind('.')]
                    if '-' in yrs:
                        # check merged file
                        yr1, yr2 = [ int(i) for i in yrs.split('-') ]
                    else:
                        # check single year file
                        yr2 = yr1 = int(yrs)
                    for yr in range(yr1, yr2 + 1):
                        if (yr >= yearstart) and (yr <= yearend):
                            if yr in hasyrs:
                                ii = hasyrs.index(yr)
                                hasyrs.remove(yr)
                                fyr = targetyrs.pop(ii)
                                targetera5[targetera5.index(fyr)] = ff

    # Select all times
    times = ['00:00', '01:00', '02:00',
             '03:00', '04:00', '05:00',
             '06:00', '07:00', '08:00',
             '09:00', '10:00', '11:00',
             '12:00', '13:00', '14:00',
             '15:00', '16:00', '17:00',
             '18:00', '19:00', '20:00',
             '21:00', '22:00', '23:00']

    # get reanalysis
    if len(hasyrs) > 0:
        today = dt.date.today()
        if len(hasyrs) == 1:
            yy = 0
            target = targetyrs[yy]
            # The time period to analyse. Valid formats:
            #    A single date as "2010-01-01"
            #    A time period as "2010-01-01/2015-12-31"
            startdate = '{:04d}-{:02d}-{:02d}'.format(hasyrs[yy], 1, 1)
            lastdate  = '{:04d}-{:02d}-{:02d}'.format(hasyrs[yy], 12, 31)
            # ERA5 has a latency of about 5 days; ERA5-Land about three
            # months. So if current year is chosen, we limit to 10 days
            # before today for ERA5 and to 100 days before today for
            # ERA5-Land.
            downyear = True
            if hasyrs[yy] == today.year:
                lastdate = today - dt.timedelta(days=latency)
                if hasyrs[yy] != lastdate.year:
                    print(f'{reanalysis_model} not available yet for this'
                          f' year')
                    downyear = False
                lastdate = lastdate.isoformat()
            if downyear:
                dates = (startdate + "/" + lastdate)
                print('Retrieve ', dates, target)
                data_file_era5 = _get_era5_single_level5(
                    vars, dates, times, area, target, reanalysis_model=rmodel,
                    output_format=output_format)
        else:
            # Do one by year because CDS allows only download of 100000 items
            # at a time.
            # Items are time steps and variables but a map is one item.
            # So 10 variables, 24 hours a day per year are: 10*24*365 = 87600.
            datestarget = []
            for yy in range(len(hasyrs)):
                target = targetyrs[yy]
                # The time period to analyse. Valid formats:
                #    A single date as "2010-01-01"
                #    A time period as "2010-01-01/2015-12-31"
                startdate = '{:04d}-{:02d}-{:02d}'.format(hasyrs[yy], 1, 1)
                lastdate  = '{:04d}-{:02d}-{:02d}'.format(hasyrs[yy], 12, 31)
                # ERA5 has a latency of about 5 days; ERA5-Land about three
                # months. So if current year is chosen, we limit to 10 days
                # before today for ERA5 and to 100 days before today for
                # ERA5-Land.
                downyear = True
                if hasyrs[yy] == today.year:
                    lastdate = today - dt.timedelta(days=latency)
                    if hasyrs[yy] != lastdate.year:
                        print(f'{reanalysis_model} not available yet for this'
                              f' year')
                        downyear = False
                    lastdate = lastdate.isoformat()
                if downyear:
                    dates = (startdate + "/" + lastdate)
                    datestarget += [(dates, target)]
                    print('Pool ', dates, target)
            from functools import partial
            getit = partial(_get_era5_single_level5_datestarget, vars, times,
                            area, reanalysis_model=rmodel,
                            output_format=output_format)
            from multiprocessing import Pool
            pool = Pool(processes=len(hasyrs))
            pool.map(getit, datestarget, None)

    return targetera5


# --------------------------------------------------------------------
# Script
#

if __name__ == "__main__":

    import argparse

    area = ''
    oformat = ''
    override = False
    path = '.'
    reanalysis_model = 'era5'
    varis = ('10m_u_component_of_wind,'
             '10m_v_component_of_wind,'
             '2m_temperature,2m_dewpoint_temperature,'
             'total_precipitation,snowfall,'
             'surface_pressure,'
             'surface_solar_radiation_downwards,'
             'surface_thermal_radiation_downwards')
    years = None

    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=('Download ERA5 or ERA5-Land data from Copernicus Climate'
                     ' Data Store'
                     ' https://climate.copernicus.eu/climate-data-store.'))
    hstr = ('area format as either lat,lon or'
            ' NorthLat/WestLon/SouthLat/EastLon, e.g. global 90/-180/-90/180,'
            ' mandatory.')
    parser.add_argument('-a', '--area', action='store', default=area,
                        dest='area', metavar='area', help=hstr)
    hstr  = ('Output format netcdf or grib. (default: netcdf if era5,'
             ' grib if era5-land).')
    parser.add_argument('-f', '--format', action='store', default=oformat,
                        dest='oformat', metavar='format', help=hstr)
    hstr = ('Do not check that output file already exists that includes'
            ' request. Override existing output file (default: False).')
    parser.add_argument('-o', '--override', action='store_true',
                        default=override, dest='override',
                        help=hstr)
    hstr = 'Output directory (default: current directory ".").'
    parser.add_argument('-p', '--path', action='store', default=path,
                        dest='path', metavar='path', help=hstr)
    hstr = ('Reanalyis model to download, either: era5, era5-land or era5land'
            ' (default: era5)')
    parser.add_argument('-r', '--reanalyis-model', action='store',
                        default=reanalysis_model, dest='reanalyis_model',
                        metavar='reanalysis_model', help=hstr)
    hstr = (f'Comma-separated variable list var1,var2,... (default: forcing'
            f' variables of ecosystem model MuSICA: {varis}.')
    parser.add_argument('-v', '--variables', action='store', default=varis,
                        dest='varis', metavar='variables', help=hstr)
    hstr = ('years format is startyear,endyear (default: 1979,current-1 for'
            ' ERA5 and 1981,current-1 for ERA5-Land.)')
    parser.add_argument('-y', '--years', action='store', default=years,
                        dest='years', metavar='years', help=hstr)

    args             = parser.parse_args()
    area             = args.area
    oformat          = args.oformat
    override         = args.override
    path             = args.path
    reanalysis_model = args.reanalyis_model
    varis            = args.varis
    years            = args.years

    del parser, args

    if not area:
        raise ValueError('area as either lat,lon or'
                         ' NorthLat/WestLon/SouthLat/EastLon must be given'
                         ' with -a option, e.g. --area="90/-180/-90/180"'
                         ' or a specific location like FR-Hes:'
                         ' -a 48.6742167,7.0646167.')

    # years
    if years:
        if ',' in years:
            syears = years.split(',')
            assert len(syears) == 2, 'years format is startyear,endyear.'
            lyears = [ int(i) for i in syears ]
        else:
            year   = int(years)
            lyears = (year, year)
    else:
        lyears = None

    # variables
    if ',' in varis:
        svars = varis.split(',')
    else:
        svars = [varis]

    import time as ptime
    t1 = ptime.time()

    era5files = get_era5(vars=svars, area=area, years=lyears, path=path,
                         override=override, reanalysis_model=reanalysis_model,
                         output_format=oformat)
    print('Files: ', era5files)

    t2    = ptime.time()
    strin = ('[m]: {:.1f}'.format((t2 - t1) / 60.) if (t2 - t1) > 60. else
             '[s]: {:d}'.format(int(t2 - t1)))
    print('Time elapsed', strin)



#!/usr/bin/env python
import numpy as np
import warnings


def itc(H, zeta, ustar, varu, varw, vart, rho, lat, limit, outdir, plot=False):
    '''
    Calculation of integral turbulence characteristics after
    Thomas & Foken (2002). Plots itc's in outdir and returns flags if values
    deviate by limit from the itc model.


    Definition
    ----------
    itc(H, zeta, ustar, varu, varw, vart, rho, lat, limit, outdir, plot=False):


    Input
    -----
    H           np.array(N), sensible heat flux [W/m2]
    zeta        np.array(N), Monin-Obukhov stability parameter (z/L) [-]
    ustar       np.array(N), friction velocity  [m/s]
    varu        np.array(N), variance of u-component [m2/s2]
    varw        np.array(N), variance of w-component [m2/s2]
    vart        np.array(N), variance of tempreature [K2]
    rho         np.array(N), air density [kg/m3]
    lat         float, latitude of tower position [dec deg]
    limit       float, limit of exceedance from the model [-]
    outdir      str, path where the plots shall be saved


    Optional Input
    --------------
    plot        bool, if True plots are generated


    Output
    ------
    itcu          np.array(N), 0 where itcu did not exceed model limits, 1 where
                  model limit is exceeded, 2 where itcu exceeds stability limits
    itcw          np.array(N), 0 where itcw did not exceed model limits, 1 where
                  model limit is exceeded, 2 where itcw exceeds stability limits
    itct          np.array(N), 0 where itct did not exceed model limits, 1 where
                  model limit is exceeded, 2 where itct exceeds stability limits
    itc_u_pos.pdf plot with itcu of positive stability range
    itc_u_neg.pdf plot with itcu of negative stability range
    itc_w_pos.pdf plot with itcw of positive stability range
    itc_w_neg.pdf plot with itcw of negative stability range
    itc_t.pdf     plot with itct


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2014 Arndt Piayda

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written,  AP, Aug 2014
    '''

    ############################################################################
    # itc models
    def itc_flag(meas, mod, limit):
        return np.where(np.logical_or(meas>mod*(1.+limit), meas<mod*(1.-limit)), 1, 0)

    def u_pos(value, f):
        return 0.44*np.log((1. * f)/value)+6.3
    def u_neg(value):
        return 4.15*np.abs(value)**(1./8.)
    def w_pos(value, f):
        return 0.21*np.log((1. * f)/value)+3.1
    def w_neg(value):
        return 1.3*(1-2*value)**(1./3.)
    def t_all(value):
        return np.where(value>0.02, 1.4*np.abs(value)**(-1./4.),\
                   np.where((0.02>=value)&(value>-0.0625),\
                       0.5*np.abs(value)**(-1./2.),\
                           np.where((-0.0625>=value)&(value>-1),\
                               np.abs(value)**(-1./4.),\
                                   np.abs(value)**(-1./3.))))

    ############################################################################
    # calculate coriolis factor
    omega = 2.*np.pi/86164.1 # anglevelocity of earth, day of 23h56min4.1sec
    phi   = lat/360.*2.*np.pi
    f     = 2*omega*np.sin(phi) # coriolis factor

    ############################################################################
    # domains
    zeta_neg = (-3.<zeta)  & (zeta<=-0.2)
    zeta_pos = (-0.2<zeta) & (zeta<0.4)

    ############################################################################
    # itcu
    itcu = np.where(np.isnan(varu), np.NaN, 0).astype(np.int)

    meas_u_pos      = np.sqrt(varu[zeta_pos])/ustar[zeta_pos]
    mod_u_pos       = u_pos(ustar[zeta_pos], f)
    itcu[zeta_pos] = itc_flag(meas_u_pos, mod_u_pos, limit)

    meas_u_neg      = np.sqrt(varu[zeta_neg])/ ustar[zeta_neg]
    mod_u_neg       = u_neg(zeta[zeta_neg])
    itcu[zeta_neg] = itc_flag(meas_u_neg, mod_u_neg, limit)

    itcu[(~np.isnan(itcu)) & (~zeta_neg) & (~zeta_pos)] = 2

    ############################################################################
    # itcw
    itcw = np.where(np.isnan(varw), np.NaN, 0).astype(np.int)

    meas_w_pos      = np.sqrt(varw[zeta_pos])/ustar[zeta_pos]
    mod_w_pos       = w_pos(ustar[zeta_pos], f)
    itcw[zeta_pos] = itc_flag(meas_w_pos, mod_w_pos, limit)

    meas_w_neg      = np.sqrt(varw[zeta_neg])/ustar[zeta_neg]
    mod_w_neg       = w_neg(zeta[zeta_neg])
    itcw[zeta_neg] = itc_flag(meas_w_neg, mod_w_neg, limit)

    itcw[(~np.isnan(itcw)) & (~zeta_neg) & (~zeta_pos)] = 2

    ############################################################################
    # itct
    itct = np.where(np.isnan(vart), np.NaN, 0).astype(np.int)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        meas_t = np.sqrt(vart) * 1004.67 * rho * ustar/np.abs(H)
    mod_t  = t_all(zeta)
    itct  = itc_flag(meas_t, mod_t, limit)

    ############################################################################
    # plots
    if plot:
        import matplotlib
        import matplotlib.pyplot as plt
        import matplotlib.backends.backend_pdf as pdf

        x_pos     = np.log((1 * f)/ustar[zeta_pos])
        pos_range = np.arange(np.nanmin(ustar[zeta_pos]),
                              np.nanmax(ustar[zeta_pos]), 0.01)
        x_mod     = np.log((1 * f)/pos_range)

        x_neg     = zeta[zeta_neg]
        neg_range = np.arange(np.nanmin(x_neg),np.nanmax(x_neg), 0.01)

        ########################################################################
        # itcu: 0.44*ln((z_+*f)/u_star)+6.3
        fig1 = plt.figure(1)
        sub1 = fig1.add_subplot(111)
        l1 = sub1.plot(x_pos, meas_u_pos, 'bo')
        l2 = sub1.plot(x_mod, u_pos(pos_range, f), 'g-',\
                       label='0.44*ln((z_+*f)/u_star)+6.3')
        l3 = sub1.plot(x_mod, u_pos(pos_range, f)*1.3, 'r-', label='+ 30%')
        l4 = sub1.plot(x_mod, u_pos(pos_range, f)*0.7, 'r-', label='- 30%')
        sub1.set_ylabel('sigma(u)/u_star')
        sub1.set_xlabel('ln(z_+*f/u_star)')
        sub1.legend(loc='best')
        plt.title('for -0.2 < zeta < 0.4')

        ########################################################################
        # itcu: 4.15*(np.abs(zeta))**(1/8)
        fig2 = plt.figure(2)
        sub1 = fig2.add_subplot(111)
        l1 = sub1.plot(x_neg, meas_u_neg, 'bo')
        l2 = sub1.plot(neg_range, u_neg(neg_range), 'g-',\
                       label='4.15*(np.abs(zeta))**(1/8)')
        l3 = sub1.plot(neg_range, u_neg(neg_range)*1.3, 'r-', label='+30%')
        l4 = sub1.plot(neg_range, u_neg(neg_range)*0.7, 'r-', label='-30%')
        sub1.set_ylabel('sigma(u)/u_star')
        sub1.set_xlabel('zeta')
        sub1.legend(loc='best')
        plt.title('for -3 < zeta <= -0.2')

        ########################################################################
        # itcw: 0.21*ln((z_+*f)/u_star)+3.1
        fig3 = plt.figure(3)
        sub1 = fig3.add_subplot(111)
        l1 = sub1.plot(x_pos, meas_w_pos, 'bo')
        l2 = sub1.plot(x_mod, w_pos(pos_range, f), 'g-',\
                       label='0.21*ln((z_+*f)/u_star)+3.1')
        l3 = sub1.plot(x_mod, w_pos(pos_range, f)*1.3, 'r-', label='+ 30%')
        l4 = sub1.plot(x_mod, w_pos(pos_range, f)*0.7, 'r-', label='- 30%')
        sub1.set_ylabel('sigma(w)/u_star')
        sub1.set_xlabel('ln(z_+*f/u_star)')
        sub1.legend(loc='best')
        plt.title('for -0.2 < zeta < 0.4')

        ########################################################################
        # itcw: 1.3*(1-2*zeta)**(1/3)
        fig4 = plt.figure(4)
        sub1 = fig4.add_subplot(111)
        l1 = sub1.plot(x_neg, meas_w_neg, 'bo')
        l2 = sub1.plot(neg_range, w_neg(neg_range), 'g-',\
                       label='1.3*(1-2*zeta)**(1/3)')
        l3 = sub1.plot(neg_range, w_neg(neg_range)*1.3, 'r-', label='+30%')
        l4 = sub1.plot(neg_range, w_neg(neg_range)*0.7, 'r-', label='-30%')
        sub1.set_ylabel('sigma(w)/u_star')
        sub1.set_xlabel('zeta')
        sub1.legend(loc='best')
        plt.title('for -3 < zeta <= -0.2')

        ########################################################################
        # itct: 1>zeta>0.02      => 1.4*np.abs(zeta)**(-1/4)
        #        0.02>zeta>-0.062 => 0.5*np.abs(zeta)**(-1/2)
        #        -0.062>zeta>-1   => 1.0*np.abs(zeta)**(-1/4)
        #        -1>zeta          => 1.0*np.abs(zeta)**(-1/3)
        fig5 = plt.figure(5)
        sub1 = fig5.add_subplot(111)
        x_mod = np.arange(-1,0.5,0.01)
        y_mod  = t_all(x_mod)
        l1 = sub1.plot(zeta, meas_t, 'bo')
        l2 = sub1.plot(x_mod, y_mod, 'g-', label='x*|zeta|**(-1/y)')
        l3 = sub1.plot(x_mod, y_mod*1.3, 'r-', label='+30%')
        l4 = sub1.plot(x_mod, y_mod*0.7, 'r-', label='-30%')
        sub1.set_ylabel('sigma(T)/T_star')
        sub1.set_xlabel('zeta')
        sub1.set_xlim(-1,0.5)
        sub1.set_ylim(0,10)
        sub1.legend(loc='best')
        plt.show()

        ########################################################################
        # save plots
        pp1  = pdf.PdfPages(outdir+'/itc_u_pos.pdf')
        pp2  = pdf.PdfPages(outdir+'/itc_u_neg.pdf')
        pp3  = pdf.PdfPages(outdir+'/itc_w_pos.pdf')
        pp4  = pdf.PdfPages(outdir+'/itc_w_neg.pdf')
        pp5  = pdf.PdfPages(outdir+'/itc_t.pdf')

        fig1.savefig(pp1, format='pdf')
        fig2.savefig(pp2, format='pdf')
        fig3.savefig(pp3, format='pdf')
        fig4.savefig(pp4, format='pdf')
        fig5.savefig(pp5, format='pdf')

        pp1.close()
        pp2.close()
        pp3.close()
        pp4.close()
        pp5.close()

    return itcu, itcw, itct

if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import numpy as np

def pi(s=None, m=None, norm=None, b=False, evalues=False, ematrix=False):
    """
        Calculates the parameter importance index PI or alternatively the B index.
        The indices can be normalised and the normalisation can be output as well,
        e.g. to calculate PI-tilde aferwards.


        Definition
        ----------
        def pi(s=s, m=m, norm=None, b=False, evalues=False, ematrix=False):


        Optional Input
        --------------
        Either
          s      sensitivity matrix[nparams,n]
        or
          m      transpose(s).s matrix[nparams,nparams]
        norm     normalised indexes; also output normalisation
                 'sum':   index /= sum(index)
                 'ev':    index /= sum(eigenvalues)
                 'evsum': both index /= sum(eigenvalues) then index /= sum(index)
        b        True:  output B index
                 False: output PI index
        evalues  True:  output eigenvalues
        ematrix  True:  output eigenmatrix


        Output
        ------
        Indices and normalisation, optionally eigenvalues and eigenmatrix


        Restrictions
        ------------
        Be aware that s is (Nparams x N) matrix and Python is column-major so that
        m is s.transpose(s) (not transpose(s).s).


        References
        ----------
        Goehler M, J Mai, and M Cuntz (2013)
            Use of eigendecomposition in a parameter sensitivity analysis of the Community Land Model,
            J Geophys Res 188, 904-921, doi:10.1002/jgrg.20072


        Examples
        --------
        >>> import numpy as np
        >>> s = np.array([[ 0.000630519,  0.000699618,  0.001640544,  0.000429059,  0.000460696],
        ...           [ 0.000181144,  0.000200996,  0.000477661,  0.000109410,  0.000117478],
        ...           [-0.000048173, -0.000053453, -0.000126809, -0.000031822, -0.000034169],
        ...           [ 0.000032940,  0.000036550, -0.000170568,  0.000021916,  0.000023532],
        ...           [ 0.000000000,  0.000101740,  0.000000000,  0.000000000,  0.000068367],
        ...           [-0.000020657, -0.000136825, -0.000054175, -0.000014864, -0.000092501],
        ...           [ 0.000025603,  0.000722743,  0.000067146,  0.000018423,  0.000486356],
        ...           [-0.000006285, -0.000191060, -0.000016484, -0.000004523, -0.000128557],
        ...           [-0.000099740, -0.000110671, -0.000003137, -0.000081818, -0.000087852],
        ...           [ 0.000000015,  0.000000017, -0.000000186,  0.000000006,  0.000000006]])
        >>> pi1 = pi(s)
        >>> from autostring import astr
        >>> print(astr(pi1[0:8],3,pp=True))
        ['4.407e-06' '1.270e-06' '3.390e-07' '3.135e-07' '2.170e-07' '3.793e-07' '1.650e-06' '4.342e-07']

        >>> pi2 = pi(s=s)
        >>> print(astr(pi2[0:8],3,pp=True))
        ['4.407e-06' '1.270e-06' '3.390e-07' '3.135e-07' '2.170e-07' '3.793e-07' '1.650e-06' '4.342e-07']

        >>> pi3 = pi(m=np.dot(s,np.transpose(s)))
        >>> print(astr(pi3[0:8],3,pp=True))
        ['4.407e-06' '1.270e-06' '3.390e-07' '3.135e-07' '2.170e-07' '3.793e-07' '1.650e-06' '4.342e-07']

        >>> pi4, ii4 = pi(s, norm='ev')
        >>> print(astr(pi4[0:8],3,pp=True))
        ['0.838' '0.242' '0.064' '0.060' '0.041' '0.072' '0.314' '0.083']
        >>> print(astr(ii4,3,pp=True))
        5.258e-06

        >>> pi5, ii5 = pi(s, norm='sum')
        >>> print(astr(pi5[0:8],3,pp=True))
        ['0.471' '0.136' '0.036' '0.033' '0.023' '0.041' '0.176' '0.046']
        >>> print(astr(ii5,3,pp=True))
        9.364e-06

        >>> pi6, ii6 = pi(s, norm='evsum')
        >>> print(astr(pi6[0:8],3,pp=True))
        ['0.471' '0.136' '0.036' '0.033' '0.023' '0.041' '0.176' '0.046']
        >>> print(astr(ii6,3,pp=True))
        9.364e-06

        >>> b1 = pi(s, b=True)
        >>> print(astr(b1[0:8],3,pp=True))
        ['3.975e-06' '3.271e-07' '2.344e-08' '3.255e-08' '1.503e-08' '3.086e-08' '7.644e-07' '5.336e-08']

        >>> b2, ii2 = pi(s, b=True, norm='ev')
        >>> print(astr(b2[0:8],3,pp=True))
        ['0.756' '0.062' '0.004' '0.006' '0.003' '0.006' '0.145' '0.010']

        >>> b3, ii3 = pi(s, b=True, norm='sum')
        >>> print(astr(b3[0:8],3,pp=True))
        ['0.756' '0.062' '0.004' '0.006' '0.003' '0.006' '0.145' '0.010']

        >>> b4, ii4 = pi(s, b=True, norm='evsum')
        >>> print(astr(b4[0:8],3,pp=True))
        ['0.756' '0.062' '0.004' '0.006' '0.003' '0.006' '0.145' '0.010']


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2012-2014 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, May 2012
        Modified, MC, Feb 2013 - ported to Python 3
                  MC, Apr 2014 - assert
    """
    # Check input
    assert not ((s is None) and (m is None)), 'No input matrix given.'
    assert not ((s is not None) and (m is not None)), 'Only s or m matrix possible.'
    if (norm is not None):
        if norm.lower() not in ['sum','ev','evsum']:
            raise ValueError("norm must be None, 'sum', 'ev', or 'evsum'.")
    #
    if (s is not None):
      # check if s is masked array
      if type(s) == type(np.ones(1)):
         m  = np.dot(s,np.transpose(s))
      else:
         m  = np.ma.dot(s,np.transpose(s))
    mm  = m.shape
    assert mm[0] == mm[1], 'm matrix not NxN.'
    inn = mm[0]
    #
    ind = np.empty(inn)
    if b:
        for i in range(inn):
            ind[i] = m[i,i]
    else:
        import scipy.linalg as la
        ev, em = la.eigh(m)

        for i in range(inn):
            ind[i] = np.sum(np.abs(ev[:]) * np.abs(em[i,:]))

    if norm is not None:
        if norm.lower() == 'sum':
            inorm = np.sum(ind)
            ind *= 1./inorm
        elif norm.lower() == 'ev':
            if b:
                inorm = np.trace(m)
            else:
                inorm = np.sum(np.abs(ev[:]))
            ind *= 1./inorm
        elif norm.lower() == 'evsum':
            if b:
                inorm = np.trace(m)
            else:
                inorm = np.sum(np.abs(ev[:]))
            ind *= 1./inorm
            inorm2 = np.sum(ind)
            ind *= 1./inorm2
            inorm *= inorm2
        out = [ind, inorm]
        if evalues:
            out += [ev]
        if ematrix:
            out += [em]
    else:
        if evalues | ematrix:
            out = [ind]
            if evalues:
                out += [ev]
            if ematrix:
                out += [em]
        else:
            out = ind
    return out


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)
    # s = np.array([[ 0.000630519,  0.000699618,  0.001640544,  0.000429059,  0.000460696],
    #               [ 0.000181144,  0.000200996,  0.000477661,  0.000109410,  0.000117478],
    #               [-0.000048173, -0.000053453, -0.000126809, -0.000031822, -0.000034169],
    #               [ 0.000032940,  0.000036550, -0.000170568,  0.000021916,  0.000023532],
    #               [ 0.000000000,  0.000101740,  0.000000000,  0.000000000,  0.000068367],
    #               [-0.000020657, -0.000136825, -0.000054175, -0.000014864, -0.000092501],
    #               [ 0.000025603,  0.000722743,  0.000067146,  0.000018423,  0.000486356],
    #               [-0.000006285, -0.000191060, -0.000016484, -0.000004523, -0.000128557],
    #               [-0.000099740, -0.000110671, -0.000003137, -0.000081818, -0.000087852],
    #               [ 0.000000015,  0.000000017, -0.000000186,  0.000000006,  0.000000006]])
    # pi1 = pi(s)
    # print pi1
    # pi2 = pi(s=s)
    # print pi2
    # pi3 = pi(m=np.dot(s,np.transpose(s)))
    # print pi3
    # pi4, ii4 = pi(s, norm='ev')
    # print pi4, ii4
    # pi5, ii5 = pi(s, norm='sum')
    # print pi5, ii5
    # pi6, ii6 = pi(s, norm='evsum')
    # print pi6, ii6
    # b1 = pi(s, b=True)
    # print b1
    # b2, ii2 = pi(s, b=True, norm='ev')
    # print b2
    # b3, ii3 = pi(s, b=True, norm='sum')
    # print b3
    # b4, ii4 = pi(s, b=True, norm='evsum')
    # print b4




#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import numpy as np
import scipy.stats as s
# import pdb


__all__ = ['outlier']


def outlier(y, alpha=0.01, k=-1, quiet=True):
    """
    Estimates the number of outliers in an approximately normal distributed
    sample according to Rossner''s (1983) generalized Extreme Studentized
    Deviate (ESD). It returns the indeces of the outliers.


    Definition
    ----------
    def outlier(y, alpha=0.01, k=-1, quiet=True):


    Input
    -----
    y            Approx. normal distributed data


    Options
    -------
    alpha        Significance level (default: 0.01)
    k            Numbers of outliers to be checked (default: size(data)/2)
    quiet        True: Perform quietly (default: True)


    Output
    ------
    Indices of outliers in input array


    Restrictions
    ------------
    Assumptions: The input sample is assumed to be approximately
                 normal distributed


    References
    ----------
    Rosner B, Percentage Points for a Generalized ESD Many-Outlier Procedure,
        Technometrics, 25(2), 165-172, 1983
    http://www.itl.nist.gov/div898/handbook/eda/section3/eda35h3.htm (Jul 2010)
    Yu RC, Teh HW, Jaques PA, Sioutas C, Froines JR,
        Quality control of semi-continuous mobility size-fractionated
        particle number concentration data; atmospheric Environment 38,
        3341- 3348, 2004


    Examples
    --------
    >>> import numpy as np
    >>> y = np.array([-0.25,0.68,0.94,1.15,2.26,2.35,2.37,2.40,2.47,2.54,2.62,
    ...               2.64,2.90,2.92,2.92,2.93,3.21,3.26,3.30,3.59,3.68,4.30,
    ...               4.64,5.34,5.42,8.01], dtype=float)
    >>> print(outlier(y,0.05,2,quiet=False))
    # of outliers
    1
    outliers
    [8.01]
    indeces
    [25]
    [25]
    >>> #
    >>> # Example from Rosner paper
    >>> y = np.array([-0.25,0.68,0.94,1.15,1.20,1.26,1.26,1.34,1.38,1.43,1.49,
    ...               1.49,1.55,1.56,1.58,1.65,1.69,1.70,1.76,1.77,1.81,1.91,
    ...               1.94,1.96,1.99,2.06,2.09,2.10,2.14,2.15,2.23,2.24,2.26,
    ...               2.35,2.37,2.40,2.47,2.54,2.62,2.64,2.90,2.92,2.92,2.93,
    ...               3.21,3.26,3.30,3.59,3.68,4.30,4.64,5.34,5.42,6.01],
    ...               dtype=float)
    >>> print(outlier(y,0.05,10,quiet=False))
    # of outliers
    3
    outliers
    [5.34 5.42 6.01]
    indeces
    [51 52 53]
    [51 52 53]
    >>> print(outlier(y,0.05,quiet=False))
    # of outliers
    3
    outliers
    [5.34 5.42 6.01]
    indeces
    [51 52 53]
    [51 52 53]
    >>> print(outlier(y,quiet=False))
    outlier: Found no outliers.
    [-1]


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python
    library, Department of Computational Hydrosystems, Helmholtz Centre for
    Environmental Research - UFZ, Leipzig, Germany.

    Copyright (c) 2010-2021 Maren Goehler, Matthias Cuntz - mc (at) macu (dot)
    de

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.


    History
    -------
    Written,  Maren Goehler,  Jul 2010
    Modified, Matthias Cuntz, Mar 2011
                  - formulation was wrong: after removal of a point, the
                    absolute deviation was not recalculated with the new mean
                    it gives exactly the same as the example in the above
                    NIST-webpage
                  - calc mean and standard deviation with cummulative formula
                    so that there is only three sums of whole array
                  - include case with no outliers: returns -1.
           Matthias Cuntz & Maren Goehler, Oct 2011
                  - break loop if new variance < 0 in iteration
           Matthias Cuntz, Feb 2013 - ported to Python 3
           Matthias Cuntz, Oct 2013 - rossner
           Matthias Cuntz, Apr 2014 - assert
           Matthias Cuntz, Sep 2021 - code refactoring
    """
    # check user input
    assert ((alpha > 0) and (alpha <= 1)), 'alpha must between 0 and 1.'

    # Student''s t-distribution
    nn    = np.ma.count(y)
    if k == -1:
        k = nn//2
    if k > nn//2:
        print('WARNING: k > size(data)/2 -> set k=size(data)/2')
        k = nn-1
    out      = 0
    outlrs   = np.zeros(k, dtype=float)
    iioutlrs = np.zeros(k, dtype=int)
    i     = np.arange(k, dtype=int) + 1
    nnip1 = np.array(nn-i+1, dtype=float)
    nnim1 = np.array(nn-i-1, dtype=float)
    pcrit = 1.-(alpha/(2.*nnip1))
    t     = s.t.ppf(pcrit, nnim1)
    lamb  = ((nnim1+1.)*t)/np.sqrt((nnim1+t*t)*nnip1)

    # Calc mean and stddev ourselves (2-pass formula) so that we understand
    # the cummulative way to calculate mean and stddev below
    # mm = np.ma.mean(y[0:nn])
    # ss  = np.std(y[0:nn],ddof=1)
    mm  = np.ma.sum(y)/nn
    dev = y-mm
    ep  = np.ma.sum(dev)
    tt  = np.ma.sum(dev*dev)
    ss  = np.ma.sqrt((tt-ep/nn)/(nn-1))

    adev        = np.ma.abs(dev)
    thisii      = np.ma.argmax(adev)
    iioutlrs[0] = thisii
    outlrs[0]   = y[thisii]
    # normalised deviation R
    R   = (np.ma.abs(y[thisii]-mm)) / ss
    # Test R against t-distribution
    if R > lamb[0]:
        out  = 1

    if k > 1:
        for i in range(k-1):
            # remove the highest value from above and redo the procedure
            mm1 = mm
            # Cummulative way to calculate mean and stddev
            mm  = (mm*(nn-i) - y[thisii])/(nn-1-i)
            ep  = ep + mm1*(nn-i) - y[thisii] - mm*(nn-1-i)
            tt  = tt - y[thisii]*y[thisii] + (nn-i)*mm1*mm1 - (nn-1-i)*mm*mm
            if ((tt-ep/(nn-1-i)) < 0.):
                break
            ss     = np.sqrt((tt-ep/(nn-1-i))/(nn-1-i-1))
            adev   = np.ma.masked_array(np.ma.abs(y-mm),
                                        mask=np.ma.mask_or(adev.mask,
                                                           y == y[thisii]))
            thisii = np.ma.argmax(adev)
            iioutlrs[i+1] = thisii
            outlrs[i+1]   = y[thisii]
            R  = (np.abs(y[thisii]-mm)) / ss
            if R > lamb[i+1]:
                out = i+2

    if out > 0:
        # report results
        iiout = iioutlrs[0:out]
        iiout = iiout[::-1]  # reverse
        if (not quiet):
            print('# of outliers')
            print(out)
            print('outliers')
            print(y[iiout])
            print('indeces')
            print(iiout)
        return iiout
    else:
        if (not quiet):
            print('outlier: Found no outliers.')
        return np.array([-1])


def rossner(*args, **kwargs):
    """
    Wrapper for outlier
    def outlier(y, alpha=0.01, k=-1, quiet=True):


    Examples
    --------
    >>> import numpy as np
    >>> y = np.array([-0.25,0.68,0.94,1.15,2.26,2.35,2.37,2.40,2.47,2.54,2.62,
    ...               2.64,2.90,2.92,2.92,2.93,3.21,3.26,3.30,3.59,3.68,4.30,
    ...               4.64,5.34,5.42,8.01], dtype=float)
    >>> print(rossner(y,0.05,2,quiet=False))
    # of outliers
    1
    outliers
    [8.01]
    indeces
    [25]
    [25]
    >>> #
    >>> # Example from Rosner paper
    >>> y = np.array([-0.25,0.68,0.94,1.15,1.20,1.26,1.26,1.34,1.38,1.43,1.49,
    ...               1.49,1.55,1.56,1.58,1.65,1.69,1.70,1.76,1.77,1.81,1.91,
    ...               1.94,1.96,1.99,2.06,2.09,2.10,2.14,2.15,2.23,2.24,2.26,
    ...               2.35,2.37,2.40,2.47,2.54,2.62,2.64,2.90,2.92,2.92,2.93,
    ...               3.21,3.26,3.30,3.59,3.68,4.30,4.64,5.34,5.42,6.01],
    ...              dtype=float)
    >>> print(rossner(y,0.05,10,quiet=False))
    # of outliers
    3
    outliers
    [5.34 5.42 6.01]
    indeces
    [51 52 53]
    [51 52 53]
    >>> print(rossner(y,0.05,quiet=False))
    # of outliers
    3
    outliers
    [5.34 5.42 6.01]
    indeces
    [51 52 53]
    [51 52 53]
    >>> print(rossner(y,quiet=False))
    outlier: Found no outliers.
    [-1]
    """
    return outlier(*args, **kwargs)


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
import numpy as np
from pyjams.jams.date2dec   import date2dec
from pyjams.jams.dec2date   import dec2date
from pyjams.jams.autostring import astr
from pyjams.jams.eddybox    import itc
from pyjams.jams.eddybox    import spikeflag
from pyjams.jams.eddybox    import ustarflag


def fluxflag(fluxfile, metfile, outdir, swdr, T, lat, delimiter=[',',','],
             skiprows=[1,1], format=['ascii','ascii'], limit=0.3,
             analyzer='LI7000', spike_f=True, itc_f=True, ustar_f=True,
             novalue=-9999, plot=False):
    '''
    Quality flag calculation for Eddy Covariance data originating from EddyFlux.
    Including stationarity, exceeding limits, exceeding change rates, exceeding
    variances, integral turbulence characteristics, spike detection and ustar
    thresholds. Missing values are flagged with 2. Ustar thresholding works
    ONLY for a data set of ONE FULL year.


    Definition
    ----------
    fluxflag(fluxfile, metfile, outdir, Rg, T, lat, delimiter=[',',','],
             skiprows=[1,1], format=['ascii','ascii'], limit=0.3,
             analyzer='LI7000', spike_f=True, itc_f=True, ustar_f=True,
             novalue=-9999, plot=False):


    Input
    -----
    fluxfile    str, path and file name of EddyFlux output file
    metfile     str, path and file name of the meteorology file (must be in
                sync with fluxfile)
    outdir      str, path of the output folder
    swdr        int, column number of short wave incoming radiation in metfile,
                column number starts with 0 which is first data column.
                Used for swdr>0=isday
    T           int, column number of air temperature in metfile, column number
                starts with 0 which is first data column. Used for ustarflag
    lat         latitude of tower position in decimal degrees. Used for itcflag


    Optional Input
    --------------
    delimiter   list of str, delimiters of fluxfile and metfile
                (default: [',',','])
    skiprows    list of int, lines to skip at the beginning of fluxfile and
                metfile, e.g. header lines (default: [1,1])
    format      list of str, time formats of fluxfile and metfile, 'ascii' and
                'eng' possible (default: ['ascii','ascii'])
    limit       float, relative deviation limit from the itc model above which
                values are flagged with itcflag. e.g. 0.3 means 30% above and
                30% below the model (default: 0.3)
    analyzer    str, analyzer type. 'LI7000' and 'LI7500' possible. Used for
                exceeding variance limits. (default: 'LI7000')
    spike_f     bool, if True spike detection and flagging is performed
    itc_f       bool, if True itc calculation and flagging is performed
    ustar_f     bool, if True ustar threshold calculation and flagging is
                performed
    novalue     int/float, missing value of fluxfile and metfile
                (default: -9999)
    plot        bool, if True subroutines (itc, spike and ustar) perform
                plotting (default: False)


    Output
    ------
    flags.csv     file containing all calculated flags
    fluxflags.csv file containing fluxes with the sum of all respective flags
                  for each flux. Where flag>2, flag=2.
    flags.log     file containing statistics for each flag


    Restrictions
    ------------
    - ustar flagging works ONLY for a data set of ONE FULL year
    - ustar flagging works ONLY for half hourly time steps


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2014 Arndt Piayda

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written,  AP, Aug 2014
    '''
    ###########################################################################
    # reading input files
    d = np.loadtxt(fluxfile, dtype='|S100', delimiter=delimiter[0], skiprows=skiprows[0])
    m = np.loadtxt(metfile,  dtype='|S100', delimiter=delimiter[1], skiprows=skiprows[1])

    if format[0]=='ascii':
        datev   = date2dec(ascii=d[:,0])
    elif format[0]=='eng':
        datev   = date2dec(eng=d[:,0])
    else:
        raise ValueError('fluxflag: unknown format')
    if format[1]=='ascii':
        datem   = date2dec(ascii=m[:,0])
    elif format[1]=='eng':
        datem   = date2dec(eng=m[:,0])
    else:
        raise ValueError('fluxflag: unknown format')

    val = np.where(d[:,1:]=='', str(novalue), d[:,1:]).astype(np.float)
    val = np.where(val==novalue, np.NaN, val)
    met = np.where(m[:,1:]=='', str(novalue), m[:,1:]).astype(np.float)
    met = np.where(met==novalue, np.NaN, met)
    print('LOADING FILES COMPLETED')

    ###########################################################################
    # calculate isday
    isday = met[:,swdr]>0.
    fluxes = val[:,[28,29,30,31,5]] # corr. H, corr. LE, corr. E, corr. C, tau
    zeta  = val[:,26]
    varu  = val[:,41]
    ustar = val[:,4]
    varw  = val[:,8]
    vart  = val[:,9]
    rho   = val[:,62]
    H     = val[:,28]
    T     = met[:,T]

    ###########################################################################
    # calculate standard flags
    sttest2  = np.where(np.isnan(val[:,37:38]), np.NaN,
                        np.unpackbits(val[:,37:38].astype('uint8'), axis=1))
    h2oexli  = np.where(np.isnan(val[:,56]), np.NaN,
                        np.where(val[:,56] > 100, 2, 0))
    c02exli  = np.where(np.isnan(val[:,55]), np.NaN,
                        np.where(val[:,55] > 100, 2, 0))
    h2oexcr  = np.where(np.isnan(val[:,50]), np.NaN,
                        np.where(val[:,50] > 50, 2, 0))
    co2excr  = np.where(np.isnan(val[:,49]), np.NaN,
                        np.where(val[:,49] > 50, 2, 0))
    texcr    = np.where(np.isnan(val[:,48]), np.NaN,
                        np.where(val[:,48] > 50, 2, 0))
    tauexcr  = np.where(np.isnan(val[:,45]) ^ np.isnan(val[:,47]),
                        np.NaN, np.where((val[:,45]>50) ^ (val[:,47]>50), 2, 0))
    zetaexcr = np.where(np.isnan(val[:,45]) ^ np.isnan(val[:,47]) ^ np.isnan(val[:,48]),
                        np.NaN, np.where((val[:,45] > 50) ^ (val[:,47] > 50) ^ (val[:,48] > 50), 2, 0))
    wexvar   = np.where(np.isnan(val[:,8]), np.NaN,
                        np.where(val[:,8] > 3, 1, 0))
    texvar   = np.where(np.isnan(val[:,9]), np.NaN,
                        np.where(val[:,9] > 3, 1, 0))
    if analyzer=='LI7000':
        h2oexvar3  = np.where(np.isnan(val[:,11]), np.NaN,
                              np.where(val[:,11] > 3, 1, 0))
        h2oexvar10 = np.where(np.isnan(val[:,11]), np.NaN,
                              np.where(val[:,11] > 10, 1, 0))
        co2exvar3  = np.where(np.isnan(val[:,10]), np.NaN,
                              np.where(val[:,10] > 3, 1, 0))
    elif analyzer=='LI7500':
        h2oexvar3  = np.where(np.isnan(val[:,11]), np.NaN,
                              np.where(val[:,11] > 5000, 1, 0))
        h2oexvar10 = np.where(np.isnan(val[:,11]), np.NaN,
                              np.where(val[:,11] > 17300, 1, 0))
        co2exvar3  = np.where(np.isnan(val[:,10]), np.NaN,
                              np.where(val[:,10] > 3.5, 1, 0))
    else:
        raise ValueError('fluxflag: unknown analyzer')
    wvar0    = np.where(np.isnan(val[:,8]), np.NaN,
                        np.where(val[:,8] < 0.00001, 2, 0))
    h2ovar0  = np.where(np.isnan(val[:,11]), np.NaN,
                        np.where(val[:,11] < 0.00001, 2, 0))
    co2var0  = np.where(np.isnan(val[:,10]), np.NaN,
                        np.where(val[:,10] < 0.00001, 2, 0))
    print('CALCULATE STANDARD FLAGS COMPLETE')

    ###########################################################################
    # itc flag calculation
    if itc_f:
        itcu, itcw, itct = itc.itc(H, zeta, ustar, varu, varw, vart, rho, lat,
                               limit, '%s/itc'%outdir, plot=plot)
        print('ITC CALCUATION COMPLETED')
    else:
        itcu = np.zeros_like(H, dtype=np.int)
        itcw = np.zeros_like(H, dtype=np.int)
        itct = np.zeros_like(H, dtype=np.int)

    ###########################################################################
    # summing flags for each flux
    Hflag   = np.nansum(np.vstack((sttest2[:,6], itcw, texcr, wexvar, texvar,
                                   wvar0)).transpose(), 1).astype(int)
    Hflag[Hflag>1]=2
    LEflag  = np.nansum(np.vstack((sttest2[:,5], itcw, h2oexli, h2oexcr, wexvar,
                                   h2oexvar3, wvar0, h2ovar0)).transpose(), 1).astype(int)
    LEflag[LEflag>1]=2
    Eflag   = np.nansum(np.vstack((sttest2[:,5], itcw, h2oexli, h2oexcr, wexvar,
                                   h2oexvar3, wvar0, h2ovar0)).transpose(), 1).astype(int)
    Eflag[Eflag>1]=2
    Cflag   = np.nansum(np.vstack((sttest2[:,4], itcw, c02exli, co2excr, wexvar,
                                   h2oexvar10, wvar0, co2var0)).transpose(), 1).astype(int) #, co2exvar3
    Cflag[Cflag>1]=2
    Tauflag = np.nansum(np.vstack((sttest2[:,3], itcu, itcw, tauexcr, wexvar,
                                   wvar0)).transpose(), 1).astype(int)
    Tauflag[Tauflag>1]=2

    ###########################################################################
    # spike detection
    #inflag = np.zeros_like(fluxes, dtype=np.int)
    inflag = np.vstack((Hflag, LEflag, Eflag, Cflag, Tauflag)).transpose()
    if spike_f:
        spikef = spikeflag.spikeflag(datev, fluxes, inflag, isday, '%s/spike'%outdir, z=7,
                           deriv=1, udef=np.NaN, spike_v=2, plot=plot)
        print('SPIKE DETECTION COMPLETED')
    else:
        spikef = np.zeros_like(inflag, dtype=np.int)

    inflag += spikef

    ###########################################################################
    # ustar flagging
    if ustar_f:
        C_ustar_T             = np.vstack((fluxes[:,3], ustar, T)).transpose()
        C_ustar_T_flags       = np.isnan(C_ustar_T).astype(np.int)
        C_ustar_T_flags[:,0] += inflag[:,3]
        ustarf = ustarflag.ustarflag(datev, C_ustar_T, C_ustar_T_flags, isday,
                           '%s/ustar'%outdir, ustar_v=2, plot=plot)
        print('USTAR FLAGGING COMPLETED')
    else:
        ustarf = np.zeros_like(datev, dtype=np.int)

    ###########################################################################
    # the big gathering :-D
    header = np.array(['sttest2_H', ' sttest2_LE/E', ' sttest2_C', ' sttest2_tau',
                       ' h2oexli', ' c02exli', ' h2oexcr', ' co2excr', ' texcr',
                       ' tauexcr', ' zetaexcr', ' wexvar', ' texvar', ' h2oexvar3',
                       ' h2oexvar10', ' co2exvar3', ' wvar0', ' h2ovar0', ' co2var0',
                       ' itcu', ' itcw', ' itct', ' spike_H', ' spike_LE', ' spike_E',
                       ' spike_C', ' spike_tau', ' ustar'])
    maxlen = [len(x) for x in header]
    flags  = np.vstack((sttest2[:,6], sttest2[:,5], sttest2[:,4], sttest2[:,3],
                        h2oexli, c02exli, h2oexcr, co2excr, texcr, tauexcr,
                        zetaexcr, wexvar, texvar, h2oexvar3, h2oexvar10,
                        co2exvar3, wvar0, h2ovar0, co2var0, itcu, itcw, itct,
                        spikef[:,0], spikef[:,1], spikef[:,2], spikef[:,3],
                        spikef[:,4], ustarf)).transpose()
    flags = np.where(np.isnan(flags), 2, flags).astype(np.int)
    flags_str = np.zeros_like(flags, dtype='|S100')
    for i, item in enumerate(maxlen):
        flags_str[:,i] = astr(flags[:,i], prec=item)

    # save flag file
    output = np.hstack((d[:,0:1], flags_str))
    np.savetxt('%s/flags.csv'%outdir,
               np.vstack((np.concatenate((['            date'], header))[np.newaxis,:],
                          output)), '%s', delimiter=',')

    ###########################################################################
    # summing flags for each flux
    Hflag   = np.nansum(np.vstack((sttest2[:,6], itcw, texcr, wexvar, texvar,
                                   wvar0, spikef[:,0])).transpose(), 1).astype(int)
    Hflag[Hflag>1]=2
    LEflag  = np.nansum(np.vstack((sttest2[:,5], itcw, h2oexli, h2oexcr, wexvar,
                                   h2oexvar3, wvar0, h2ovar0, spikef[:,1])).transpose(), 1).astype(int)
    LEflag[LEflag>1]=2
    Eflag   = np.nansum(np.vstack((sttest2[:,5], itcw, h2oexli, h2oexcr, wexvar,
                                   h2oexvar3, wvar0, h2ovar0, spikef[:,2])).transpose(), 1).astype(int)
    Eflag[Eflag>1]=2
    Cflag   = np.nansum(np.vstack((sttest2[:,4], itcw, c02exli, co2excr, wexvar,
                                   h2oexvar10, co2exvar3, wvar0, co2var0, spikef[:,3],
                                   ustarf)).transpose(), 1).astype(int) ### why h20exvar10
    Cflag[Cflag>1]=2
    Tauflag = np.nansum(np.vstack((sttest2[:,3], itcu, itcw, tauexcr, wexvar,
                                   wvar0, spikef[:,4])).transpose(), 1).astype(int)
    Tauflag[Tauflag>1]=2
    header2     = np.array(['          H', '   Hflag', '         LE', '  LEflag', '          E', '   Eflag', '          C',
                            '   Cflag', '        TAU', ' TAUflag'])
    fluxes[np.isnan(fluxes)] = novalue
    fluxes_str  = np.array([['%11.5f'%x for x in y] for y in fluxes])
    fluxes_str  = np.where(fluxes_str=='%11.5f'%novalue, ' '*(11-len(str(novalue)))+str(novalue), fluxes_str)
    Hflag_str   = astr(Hflag, prec=8)
    LEflag_str  = astr(LEflag, prec=8)
    Eflag_str   = astr(Eflag, prec=8)
    Cflag_str   = astr(Cflag, prec=8)
    Tauflag_str = astr(Tauflag, prec=8)

    output = np.hstack((d[:,0:1], fluxes_str.repeat(2, axis=1)))
    output[:,2::2] = np.vstack((Hflag_str, LEflag_str, Eflag_str, Cflag_str,
                               Tauflag_str)).transpose()
    np.savetxt('%s/fluxflags.csv'%outdir,
               np.vstack((np.concatenate((['            date'], header2))[np.newaxis,:],
                          output)), '%s', delimiter=',')

    ###########################################################################
    # write log file with stats
    log = open('%s/flags.log'%outdir, 'w')
    log.write('flag, 0, 1, 2\n')
    for i, item in enumerate(header):
        hist = tuple(np.histogram(flags[:,i], [-0.5,0.5,1.5,2.5])[0])
        log.write(item.strip()+', %i, %i, %i\n'%hist)
    log.close()

if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function

def tsym(name):
    """
        Returns unicodes for common symbols.


        Definition
        ----------
        def tsym(name):


        Input
        -----
        name        Symbol name, case insensitive


        Output
        ------
        Raw unicode string


        Restrictions
        ------------
        Only knows a very limited number of names
        Use empty function to return names


        Known symbols
        ------------
        deg         Degree symbol
        degree      Degree symbol
        degc        Degree Celcius
        degreec     Degree Celcius
        degree c    Degree Celcius
        mu          Lowercase greek mu
        peclet      Peclet symbol
        permil      Permille sign
        permille    Permille sign
        per mil     Permille sign
        per mille   Permille sign


        Examples
        --------
        >>> print(tsym('deg'))
        \u00B0
        >>> print(tsym('degree'))
        \u00B0
        >>> print(tsym('degC'))
        \u2103
        >>> print(tsym('degreec'))
        \u2103
        >>> print(tsym('degree C'))
        \u2103
        >>> print(tsym('mu'))
        \u00B5
        >>> print(tsym('peclet'))
        \u2118
        >>> print(tsym('permil'))
        \u2030
        >>> print(tsym('Permil'))
        \u2030
        >>> print(tsym('permille'))
        \u2030
        >>> print(tsym('per mil'))
        \u2030
        >>> print(tsym('per mille'))
        \u2030


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2011-2013 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Jun 2011
        Modified, MC, Feb 2013 - ported to Python 3
                  MC, Mar 2013 - removed raw
    """
    #
    # Define symbol dictionary
    symdict = ({
               'deg'       : '\u00B0',
               'degree'    : '\u00B0',
               'degc'      : '\u2103',
               'degreec'   : '\u2103',
               'degree c'  : '\u2103',
               'mu'        : '\u00B5',
               'peclet'    : '\u2118',
               'permil'    : '\u2030',
               'permille'  : '\u2030',
               'per mil'   : '\u2030',
               'per mille' : '\u2030'
              })

    #
    # lookup symbol
    try:
        out = symdict[name.lower()]
    except KeyError:
        print("TSYM: Symbol not known: %s" % name)
        return None

    return out


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)
